/*
 * $Id: //depot/sw/branches/linuxsrc_5.0.3/apps/busybox-1.00-pre10/shell/athcli.h#1 $
 */
#ifndef _ATHCLI_H
#define _ATHCLI_H

#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <linux/wireless.h>
/*##include <iwlib.h>*/
#include "iwlib.h"

/*
 * CLI command definition
 */
enum {
    CLI_CMD_CHANGE = 0,
    CLI_CMD_GET,
    CLI_CMD_SET,
    CLI_CMD_SAVE,
    CLI_MAX_CMD
};

struct cli_param;

typedef int (*cli_handler)(struct cli_param *param, int param_index, void *);

struct cli_param {
    char        *name;    /* parameter name */
    u_int32_t   index;    /* parameter flags */
                          /* 0-15 sub ioctl */
                          /* 16-23 format */
                          /* 23-31 flags */
    cli_handler get;      /* parameter getter */
    cli_handler set;      /* parameter setter */
};

#define CLI_FORMAT_BOOLEAN      0x00010000
#define CLI_FORMAT_INTEGER      0x00020000
#define CLI_FORMAT_TEXT         0x00040000
#define CLI_FORMAT_IP           0x00080000
#define CLI_FORMAT_SELF         0x00100000
#define CLI_FORMAT_NULL         0x00200000

#define CLI_F_STOPIF            0x01000000 /* Interface should be stopped before change configuration */
#define CLI_F_PROBE_ONLY        0x02000000 /* Value should be read from driver instead of configuration file */
#define CLI_F_HIDDEN            0x04000000 /* Hidden commands */
#define CLI_F_ZERO_PARAM        0x08000000 /* This command has zero parameters */
#define CLI_F_RADIO             0x10000000 /* This command is per radio */
#define CLI_F_BSS               0x20000000 /* This command is per BSS (virtual AP) */

#define CLI_PARAM_SUBIOCTL(flags)   (flags & 0xffff)

/*
 * MADWIFI SUB-IOCTLS. They must be in sync with madwifi driver
 */
enum {
	IEEE80211_WMMPARAMS_CWMIN       = 1,
	IEEE80211_WMMPARAMS_CWMAX       = 2,
	IEEE80211_WMMPARAMS_AIFS       	= 3,
	IEEE80211_WMMPARAMS_TXOPLIMIT	= 4,
	IEEE80211_WMMPARAMS_ACM		= 5,
	IEEE80211_WMMPARAMS_NOACKPOLICY	= 6,	
};

enum {
	IEEE80211_PARAM_TURBO		= 1,	/* turbo mode */
	IEEE80211_PARAM_MODE		= 2,	/* phy mode (11a, 11b, etc.) */
	IEEE80211_PARAM_AUTHMODE	= 3,	/* authentication mode */
	IEEE80211_PARAM_PROTMODE	= 4,	/* 802.11g protection */
	IEEE80211_PARAM_MCASTCIPHER	= 5,	/* multicast/default cipher */
	IEEE80211_PARAM_MCASTKEYLEN	= 6,	/* multicast key length */
	IEEE80211_PARAM_UCASTCIPHERS	= 7,	/* unicast cipher suites */
	IEEE80211_PARAM_UCASTCIPHER	= 8,	/* unicast cipher */
	IEEE80211_PARAM_UCASTKEYLEN	= 9,	/* unicast key length */
	IEEE80211_PARAM_WPA		= 10,	/* WPA mode (0,1,2) */
	IEEE80211_PARAM_ROAMING		= 12,	/* roaming mode */
	IEEE80211_PARAM_PRIVACY		= 13,	/* privacy invoked */
	IEEE80211_PARAM_COUNTERMEASURES	= 14,	/* WPA/TKIP countermeasures */
	IEEE80211_PARAM_DROPUNENCRYPTED	= 15,	/* discard unencrypted frames */
	IEEE80211_PARAM_DRIVER_CAPS	= 16,	/* driver capabilities */
	IEEE80211_PARAM_MACCMD		= 17,	/* MAC ACL operation */
	IEEE80211_PARAM_WMM		= 18,	/* WMM mode (on, off) */
	IEEE80211_PARAM_HIDESSID	= 19,	/* hide SSID mode (on, off) */
	IEEE80211_PARAM_APBRIDGE	= 20,	/* AP inter-sta bridging */
	IEEE80211_PARAM_KEYMGTALGS	= 21,	/* key management algorithms */
	IEEE80211_PARAM_RSNCAPS		= 22,	/* RSN capabilities */
	IEEE80211_PARAM_INACT		= 23,	/* station inactivity timeout */
	IEEE80211_PARAM_INACT_AUTH	= 24,	/* station auth inact timeout */
	IEEE80211_PARAM_INACT_INIT	= 25,	/* station init inact timeout */
	IEEE80211_PARAM_ABOLT		= 26,	/* Atheros Adv. Capabilities */
	IEEE80211_PARAM_DTIM_PERIOD	= 28,	/* DTIM period (beacons) */
	IEEE80211_PARAM_BEACON_INTERVAL	= 29,	/* beacon interval (ms) */
	IEEE80211_PARAM_DOTH		= 30,	/* 11.h is on/off */
	IEEE80211_PARAM_PWRTARGET	= 31,	/* Current Channel Pwr Constraint */
	IEEE80211_PARAM_GENREASSOC	= 32,	/* Generate a reassociation request */
	IEEE80211_PARAM_COMPRESSION	= 33,	/* compression */
	IEEE80211_PARAM_FF		= 34,	/* fast frames support  */
	IEEE80211_PARAM_XR		= 35,	/* XR support */
	IEEE80211_PARAM_BURST		= 36,	/* burst mode */
	IEEE80211_PARAM_PUREG		= 37,	/* pure 11g (no 11b stations) */
    	IEEE80211_PARAM_AR          	= 38,   /* AR support */
	IEEE80211_PARAM_WDS		= 39,	/* Enable 4 address processing */
	IEEE80211_PARAM_BGSCAN		= 40,	/* bg scanning (on, off) */
	IEEE80211_PARAM_BGSCAN_IDLE	= 41,	/* bg scan idle threshold */
	IEEE80211_PARAM_BGSCAN_INTERVAL	= 42,	/* bg scan interval */
	IEEE80211_PARAM_COVERAGECLASS	= 44,	/* coverage class */
        IEEE80211_PARAM_COUNTRY_IE 	= 45,	/* enable/disable country ie */
	IEEE80211_PARAM_CHANBW		= 60,	/* channel spacing */
	IEEE80211_PARAM_WMM_AGGRMODE    = 61, /* WMM Aggressive Mode */
	IEEE80211_PARAM_SHORTPREAMBLE 	= 62, 	/* enable/disable short Preamble */
        IEEE80211_PARAM_BLOCKDFSCHAN 	= 63, 	/* enable/disable use of DFS channels */
};

/*
 * Private IOCTL to add/delete a virtual BSS
 */
#define	IEEE80211_IOCTL_STA_INFO	(SIOCDEVPRIVATE+6)
#define	SIOC80211IFCREATE		(SIOCDEVPRIVATE+7)
#define	SIOC80211IFDESTROY	 	(SIOCDEVPRIVATE+8)

struct ieee80211_clone_params {
	char		icp_name[IFNAMSIZ];	/* device name */
	u_int16_t	icp_opmode;		/* operating mode */
	u_int16_t	icp_flags;		/* see below */
#define	IEEE80211_CLONE_BSSID	0x0001		/* allocate unique mac/bssid */
#define	IEEE80211_NO_STABEACONS	0x0002		/* Do not setup the station beacon timers */
};

enum ieee80211_opmode {
	IEEE80211_M_STA		= 1,	/* infrastructure station */
	IEEE80211_M_IBSS 	= 0,	/* IBSS (adhoc) station */
	IEEE80211_M_AHDEMO	= 3,	/* Old lucent compatible adhoc demo */
	IEEE80211_M_HOSTAP	= 6,	/* Software Access Point */
	IEEE80211_M_MONITOR	= 8,	/* Monitor mode */
	IEEE80211_M_WDS		= 2	/* WDS link */
};

/* 
 * MAC ACL operations.
 */
enum {
	IEEE80211_MACCMD_POLICY_OPEN	= 0,	/* set policy: no ACL's */
	IEEE80211_MACCMD_POLICY_ALLOW	= 1,	/* set policy: allow traffic */
	IEEE80211_MACCMD_POLICY_DENY	= 2,	/* set policy: deny traffic */
	IEEE80211_MACCMD_FLUSH		= 3,	/* flush ACL database */
	IEEE80211_MACCMD_DETACH		= 4,	/* detach ACL policy */
	IEEE80211_MACCMD_MAX
};

struct cli_country {
    int     cc;
    char    *iso_name;
};

/*
 * Country/Region Codes from MS WINNLS.H
 * Numbering from ISO 3166
 */
enum CountryCode {
    CTRY_ALBANIA              = 8,       /* Albania */
    CTRY_ALGERIA              = 12,      /* Algeria */
    CTRY_ARGENTINA            = 32,      /* Argentina */
    CTRY_ARMENIA              = 51,      /* Armenia */
    CTRY_AUSTRALIA            = 36,      /* Australia */
    CTRY_AUSTRIA              = 40,      /* Austria */
    CTRY_AZERBAIJAN           = 31,      /* Azerbaijan */
    CTRY_BAHRAIN              = 48,      /* Bahrain */
    CTRY_BELARUS              = 112,     /* Belarus */
    CTRY_BELGIUM              = 56,      /* Belgium */
    CTRY_BELIZE               = 84,      /* Belize */
    CTRY_BOLIVIA              = 68,      /* Bolivia */
    CTRY_BOSNIA_HERZEGOWINA   = 70, 	 /* Bosnia and Herzegowina */
    CTRY_BRAZIL               = 76,      /* Brazil */
    CTRY_BRUNEI_DARUSSALAM    = 96,      /* Brunei Darussalam */
    CTRY_BULGARIA             = 100,     /* Bulgaria */
    CTRY_CANADA               = 124,     /* Canada */
    CTRY_CHILE                = 152,     /* Chile */
    CTRY_CHINA                = 156,     /* People's Republic of China */
    CTRY_COLOMBIA             = 170,     /* Colombia */
    CTRY_COSTA_RICA           = 188,     /* Costa Rica */
    CTRY_CROATIA              = 191,     /* Croatia */
    CTRY_CYPRUS               = 196,
    CTRY_CZECH                = 203,     /* Czech Republic */
    CTRY_DENMARK              = 208,     /* Denmark */
    CTRY_DOMINICAN_REPUBLIC   = 214,     /* Dominican Republic */
    CTRY_ECUADOR              = 218,     /* Ecuador */
    CTRY_EGYPT                = 818,     /* Egypt */
    CTRY_EL_SALVADOR          = 222,     /* El Salvador */
    CTRY_ESTONIA              = 233,     /* Estonia */
    CTRY_FAEROE_ISLANDS       = 234,     /* Faeroe Islands */
    CTRY_FINLAND              = 246,     /* Finland */
    CTRY_FRANCE               = 250,     /* France */
    CTRY_FRANCE2              = 255,     /* France2 */
    CTRY_GEORGIA              = 268,     /* Georgia */
    CTRY_GERMANY              = 276,     /* Germany */
    CTRY_GREECE               = 300,     /* Greece */
    CTRY_GUATEMALA            = 320,     /* Guatemala */
    CTRY_HONDURAS             = 340,     /* Honduras */
    CTRY_HONG_KONG            = 344,     /* Hong Kong S.A.R., P.R.C. */
    CTRY_HUNGARY              = 348,     /* Hungary */
    CTRY_ICELAND              = 352,     /* Iceland */
    CTRY_INDIA                = 356,     /* India */
    CTRY_INDONESIA            = 360,     /* Indonesia */
    CTRY_IRAN                 = 364,     /* Iran */
    CTRY_IRAQ                 = 368,     /* Iraq */
    CTRY_IRELAND              = 372,     /* Ireland */
    CTRY_ISRAEL               = 376,     /* Israel */
    CTRY_ITALY                = 380,     /* Italy */
    CTRY_JAMAICA              = 388,     /* Jamaica */
    CTRY_JAPAN                = 392,     /* Japan */
    CTRY_JAPAN1               = 393,     /* Japan (JP1) */
    CTRY_JAPAN2               = 394,     /* Japan (JP0) */
    CTRY_JAPAN3               = 395,     /* Japan (JP1-1) */
    CTRY_JAPAN4               = 396,     /* Japan (JE1) */
    CTRY_JAPAN5               = 397,     /* Japan (JE2) */
    CTRY_JAPAN6		          = 399,	 /* Japan (JP6) */
    CTRY_JAPAN7		          = 4007,	 /* Japan (J7) */
    CTRY_JAPAN8		          = 4008,	 /* Japan (J8) */
    CTRY_JAPAN9		          = 4009,	 /* Japan (J9) */
    CTRY_JAPAN10	          = 4010,	 /* Japan (J10) */
    CTRY_JAPAN11	          = 4011,	 /* Japan (J11) */
    CTRY_JAPAN12	          = 4012,	 /* Japan (J12) */
    CTRY_JAPAN13	          = 4013,	 /* Japan (J13) */
    CTRY_JAPAN14	          = 4014,	 /* Japan (J14) */
    CTRY_JAPAN15	          = 4015,	 /* Japan (J15) */
    CTRY_JAPAN16	          = 4016,	 /* Japan (J16) */
    CTRY_JAPAN17	          = 4017,	 /* Japan (J17) */
    CTRY_JAPAN18	          = 4018,	 /* Japan (J18) */
    CTRY_JAPAN19	          = 4019,	 /* Japan (J19) */
    CTRY_JAPAN20	          = 4020,	 /* Japan (J20) */
    CTRY_JAPAN21	          = 4021,	 /* Japan (J21) */
    CTRY_JAPAN22	          = 4022,	 /* Japan (J22) */
    CTRY_JAPAN23	          = 4023,	 /* Japan (J23) */
    CTRY_JAPAN24	          = 4024,	 /* Japan (J24) */
    CTRY_JAPAN25              = 4025,	 /* Japan (J25) */
    CTRY_JAPAN26              = 4026,	 /* Japan (J26) */
    CTRY_JAPAN27              = 4027,	 /* Japan (J27) */
    CTRY_JAPAN28              = 4028,	 /* Japan (J28) */
    CTRY_JAPAN29              = 4029,	 /* Japan (J29) */
    CTRY_JAPAN30              = 4030,	 /* Japan (J30) */
    CTRY_JAPAN31              = 4031,	 /* Japan (J31) */
    CTRY_JAPAN32              = 4032,	 /* Japan (J32) */
    CTRY_JAPAN33              = 4033,	 /* Japan (J33) */
    CTRY_JAPAN34              = 4034,	 /* Japan (J34) */
    CTRY_JAPAN35              = 4035,	 /* Japan (J35) */
    CTRY_JAPAN36              = 4036,	 /* Japan (J36) */
    CTRY_JAPAN37              = 4037,	 /* Japan (J37) */
    CTRY_JAPAN38              = 4038,	 /* Japan (J38) */
    CTRY_JAPAN39              = 4039,	 /* Japan (J39) */
    CTRY_JAPAN40              = 4040,	 /* Japan (J40) */
    CTRY_JAPAN41              = 4041,	 /* Japan (J41) */
    CTRY_JAPAN42              = 4042,	 /* Japan (J42) */
    CTRY_JAPAN43              = 4043,	 /* Japan (J43) */
    CTRY_JAPAN44              = 4044,	 /* Japan (J44) */
    CTRY_JAPAN45              = 4045,	 /* Japan (J45) */
    CTRY_JAPAN46              = 4046,	 /* Japan (J46) */
    CTRY_JAPAN47              = 4047,	 /* Japan (J47) */
    CTRY_JAPAN48              = 4048,	 /* Japan (J48) */
    CTRY_JAPAN49              = 4049,	 /* Japan (J49) */
    CTRY_JAPAN50              = 4050,	 /* Japan (J50) */
    CTRY_JAPAN51              = 4051,	 /* Japan (J51) */
    CTRY_JAPAN52              = 4052,	 /* Japan (J52) */
    CTRY_JAPAN53              = 4053,	 /* Japan (J53) */
    CTRY_JAPAN54              = 4054,	 /* Japan (J54) */
    CTRY_JORDAN               = 400,     /* Jordan */
    CTRY_KAZAKHSTAN           = 398,     /* Kazakhstan */
    CTRY_KENYA                = 404,     /* Kenya */
    CTRY_KOREA_NORTH          = 408,     /* North Korea */
    CTRY_KOREA_ROC            = 410,     /* South Korea */
    CTRY_KOREA_ROC2           = 411,     /* South Korea */
    CTRY_KOREA_ROC3           = 412,     /* South Korea */
    CTRY_KUWAIT               = 414,     /* Kuwait */
    CTRY_LATVIA               = 428,     /* Latvia */
    CTRY_LEBANON              = 422,     /* Lebanon */
    CTRY_LIBYA                = 434,     /* Libya */
    CTRY_LIECHTENSTEIN        = 438,     /* Liechtenstein */
    CTRY_LITHUANIA            = 440,     /* Lithuania */
    CTRY_LUXEMBOURG           = 442,     /* Luxembourg */
    CTRY_MACAU                = 446,     /* Macau */
    CTRY_MACEDONIA            = 807,     /* the Former Yugoslav Republic of Macedonia */
    CTRY_MALAYSIA             = 458,     /* Malaysia */
    CTRY_MALTA		      = 470,	 /* Malta */
    CTRY_MEXICO               = 484,     /* Mexico */
    CTRY_MONACO               = 492,     /* Principality of Monaco */
    CTRY_MOROCCO              = 504,     /* Morocco */
    CTRY_NETHERLANDS          = 528,     /* Netherlands */
    CTRY_NEW_ZEALAND          = 554,     /* New Zealand */
    CTRY_NICARAGUA            = 558,     /* Nicaragua */
    CTRY_NORWAY               = 578,     /* Norway */
    CTRY_OMAN                 = 512,     /* Oman */
    CTRY_PAKISTAN             = 586,     /* Islamic Republic of Pakistan */
    CTRY_PANAMA               = 591,     /* Panama */
    CTRY_PARAGUAY             = 600,     /* Paraguay */
    CTRY_PERU                 = 604,     /* Peru */
    CTRY_PHILIPPINES          = 608,     /* Republic of the Philippines */
    CTRY_POLAND               = 616,     /* Poland */
    CTRY_PORTUGAL             = 620,     /* Portugal */
    CTRY_PUERTO_RICO          = 630,     /* Puerto Rico */
    CTRY_QATAR                = 634,     /* Qatar */
    CTRY_ROMANIA              = 642,     /* Romania */
    CTRY_RUSSIA               = 643,     /* Russia */
    CTRY_SAUDI_ARABIA         = 682,     /* Saudi Arabia */
    CTRY_SINGAPORE            = 702,     /* Singapore */
    CTRY_SLOVAKIA             = 703,     /* Slovak Republic */
    CTRY_SLOVENIA             = 705,     /* Slovenia */
    CTRY_SOUTH_AFRICA         = 710,     /* South Africa */
    CTRY_SPAIN                = 724,     /* Spain */
    CTRY_SRI_LANKA            = 144,     /* Spain */
    CTRY_SWEDEN               = 752,     /* Sweden */
    CTRY_SWITZERLAND          = 756,     /* Switzerland */
    CTRY_SYRIA                = 760,     /* Syria */
    CTRY_TAIWAN               = 158,     /* Taiwan */
    CTRY_THAILAND             = 764,     /* Thailand */
    CTRY_TRINIDAD_Y_TOBAGO    = 780,     /* Trinidad y Tobago */
    CTRY_TUNISIA              = 788,     /* Tunisia */
    CTRY_TURKEY               = 792,     /* Turkey */
    CTRY_UAE                  = 784,     /* U.A.E. */
    CTRY_UKRAINE              = 804,     /* Ukraine */
    CTRY_UNITED_KINGDOM       = 826,     /* United Kingdom */
    CTRY_UNITED_STATES        = 840,     /* United States */
    CTRY_UNITED_STATES_FCC49  = 842,     /* United States (Public Safety)*/
    CTRY_URUGUAY              = 858,     /* Uruguay */
    CTRY_UZBEKISTAN           = 860,     /* Uzbekistan */
    CTRY_VENEZUELA            = 862,     /* Venezuela */
    CTRY_VIET_NAM             = 704,     /* Viet Nam */
    CTRY_YEMEN                = 887,     /* Yemen */
    CTRY_ZIMBABWE             = 716      /* Zimbabwe */
};

#endif /* end of _ATHCLI_H */
