/*
 * Copyright (c) 2005-2006 Atheros Communications, Inc.
 * All rights reserved.
 *
 */

#ifndef _DFS_H_
#define _DFS_H_

/*
 * Linux specific locking macros
 */

#define	ATH_DFSQ_LOCK(_dfs)	spin_lock_irqsave(&(_dfs)->dfs_radarqlock, (_dfs)->dfs_radarqlockflags)
#define	ATH_DFSQ_UNLOCK(_dfs)	spin_unlock_irqrestore(&(_dfs)->dfs_radarqlock, (_dfs)->dfs_radarqlockflags)
#define	ATH_DFSQ_LOCK_INIT(_dfs)	spin_lock_init(&(_dfs)->dfs_radarqlock)

#define	ATH_ARQ_LOCK(_dfs)	spin_lock_irqsave(&(_dfs)->dfs_arqlock, (_dfs)->dfs_arqlockflags)
#define	ATH_ARQ_UNLOCK(_dfs)	spin_unlock_irqrestore(&(_dfs)->dfs_arqlock, (_dfs)->dfs_arqlockflags)
#define	ATH_ARQ_LOCK_INIT(_dfs)	spin_lock_init(&(_dfs)->dfs_arqlock)

#define	ATH_DFSEVENTQ_LOCK(_dfs)	spin_lock_irqsave(&(_dfs)->dfs_eventqlock, (_dfs)->dfs_eventqlockflags)
#define	ATH_DFSEVENTQ_UNLOCK(_dfs)	spin_unlock_irqrestore(&(_dfs)->dfs_eventqlock, (_dfs)->dfs_eventqlockflags)
#define	ATH_DFSEVENTQ_LOCK_INIT(_dfs)	spin_lock_init(&(_dfs)->dfs_eventqlock)


/* 
 * Default 5212/5312 radar phy parameters
 */
#define AR5212_DFS_FIRPWR	-41
#define AR5212_DFS_RRSSI	12
#define AR5212_DFS_HEIGHT	20
#define AR5212_DFS_PRSSI	22
#define AR5212_DFS_INBAND	6

/*
 * Default 5413/5416 radar phy parameters
 */
#define AR5413_DFS_FIRPWR	-44
#define AR5413_DFS_RRSSI	20
#define AR5413_DFS_HEIGHT	10
#define AR5413_DFS_PRSSI	15
#define AR5413_DFS_INBAND	6
#define AR5413_DFS_RELPWR    	8
#define AR5413_DFS_RELSTEP    	31
#define AR5413_DFS_MAXLEN   	255


#define DFS_TSMASK		0x7FFF		/* Mask for time stamp from descriptor */
#define	DFS_TSSHIFT		15		/* Shift for time stamp from descriptor */
#define	DFS_TSF_WRAP		0xFFFFFFFFFFFFFFFFULL	/* 64 bit TSF wrap value */
#define	DFS_64BIT_TSFMASK	0x0000000000007FFFULL	/* TS mask for 64 bit value */


#define	DFS_AR_RADAR_RSSI_THR		5	/* in dB */
#define	DFS_AR_RADAR_RESET_INT		1	/* in secs */
#define	DFS_AR_RADAR_MAX_HISTORY	500
#define	DFS_AR_REGION_WIDTH		128
#define	DFS_AR_RSSI_THRESH_STRONG_PKTS	17	/* in dB */
#define	DFS_AR_RSSI_DOUBLE_THRESHOLD	15	/* in dB */
#define	DFS_AR_MAX_NUM_ACK_REGIONS	9
#define	DFS_AR_ACK_DETECT_PAR_THRESH	20
#define	DFS_AR_PKT_COUNT_THRESH		20

#define	DFS_MAX_DL_SIZE			64
#define	DFS_MAX_DL_MASK			0x3F

#define DFS_NOL_TIME			30*60*1000000	/* 30 minutes in usecs */

#define DFS_WAIT_TIME			60*1000000	/* 1 minute in usecs */

#define	DFS_DISABLE_TIME		3*60*1000000	/* 3 minutes in usecs */

//#define	DFS_PARAM_FIRPWR		1
//#define	DFS_PARAM_RRSSI			2
//#define	DFS_PARAM_HEIGHT		3
//#define	DFS_PARAM_PRSSI			4
//#define	DFS_PARAM_INBAND		5
//#define	DFS_PARAM_NOL			6
//5413 specific parameters
//#define	DFS_PARAM_RELPWR		7
//#define	DFS_PARAM_RELSTEP		8
//#define DFS_PARAM_MAXLEN                9


#define	DFS_MAX_B5_SIZE			1024
#define	DFS_MAX_B5_MASK			0x000003FF	/* 1024 */

#define	DFS_MAX_RADAR_OVERLAP		6		/* Max number of overlapping filters */

#define	DFS_MAX_EVENTS			1024		/* Max number of dfs events which can be q'd */

#define DFS_RADAR_EN		0x80000000	/* Radar detect is capable */
#define DFS_AR_EN		0x40000000	/* AR detect is capable */
#define	DFS_MAX_RSSI_VALUE	0x7fffffff	/* Max rssi value */

#define DFS_BIN_MAX_PULSES      	60	/* max num of pulses in a burst */
#define DFS_BIN5_PRI_LOWER_LIMIT	990	/* us */
#define DFS_BIN5_PRI_HIGHER_LIMIT	2010	/* us */
#define DFS_BIN5_WIDTH_MARGIN	    	4	/* us */
#define DFS_BIN5_WIDTH_MARGIN_5413    	8	/* us */
#define DFS_BIN5_RSSI_MARGIN	    	5	/* dBm */
/*Following threshold is not specified but should be okay statistically*/
#define DFS_BIN5_BRI_LOWER_LIMIT	300000  /* us */

typedef	spinlock_t dfsq_lock_t;

struct dfs_pulse {
	u_int32_t	rp_numpulses;	/* Num of pulses in radar burst */
	u_int32_t	rp_pritype;	/* repetition type, fixed or variable */
	u_int32_t	rp_pulsefreq;	/* Frequency of pulses in burst */
	u_int32_t	rp_pulsevar;	/* Time variation of pulse duration for
					   matched filter (single-sided) in usecs */
	u_int32_t	rp_threshold;	/* Thershold for MF output to indicate 
					   radar match */
	u_int32_t	rp_mindur;	/* Min pulse duration to be considered for
					   this pulse type */
	u_int32_t	rp_maxdur;	/* Max pusle duration to be considered for
					   this pulse type */
	u_int32_t	rp_rssithresh;	/* Minimum rssi to be considered a radar pulse */
	u_int32_t	rp_meanoffset;	/* Offset for timing adjustment */
	int32_t		rp_rssimargin;  /* rssi threshold margin. In Turbo Mode HW reports rssi 3dBm */
                                        /* lower than in non TURBO mode. This will be used to offset that diff.*/
	u_int32_t	rp_pulseid;	/* Unique ID for identifying filter */

};

struct dfs_event {
	u_int16_t	re_ts;		/* Original 15 bit recv timestamp */
	u_int64_t	re_full_ts;	/* 64-bit full timestamp from interrupt time */
	u_int8_t	re_rssi;	/* rssi of radar event */
	u_int8_t	re_dur;		/* duration of radar pulse */
	u_int8_t	re_chanindex;	/* Channel of event */
	STAILQ_ENTRY(dfs_event)	re_list;	/* List of radar events */
};

#define DFS_AR_MAX_ACK_RADAR_DUR	511
#define DFS_AR_MAX_NUM_PEAKS		3
#define DFS_AR_ARQ_SIZE			2048	/* 8K AR events for buffer size */
#define DFS_AR_ARQ_SEQSIZE		2049	/* Sequence counter wrap for AR */

#define DFS_RADARQ_SIZE		512		/* 1K radar events for buffer size */
#define DFS_RADARQ_SEQSIZE	513		/* Sequence counter wrap for radar */
#define DFS_NUM_RADAR_STATES	64		/* Number of radar channels we keep state for */
#define DFS_MAX_NUM_RADAR_FILTERS 10		/* Max number radar filters for each type */ 
#define DFS_MAX_RADAR_TYPES	32		/* Number of different radar types */
#define DFS_MAX_PULSE_BUFFER_SIZE 1024          /* Max number of pulses kept in buffer */
#define DFS_MAX_PULSE_BUFFER_MASK 0x3ff  
struct dfs_ar_state {
	u_int16_t	ar_prevtimestamp;
	u_int32_t	ar_prevwidth;
	u_int32_t	ar_phyerrcount[DFS_AR_MAX_ACK_RADAR_DUR];
	u_int32_t	ar_acksum;
	u_int16_t	ar_peaklist[DFS_AR_MAX_NUM_PEAKS];
	u_int32_t	ar_packetthreshold;	/* Thresh to determine traffic load */
	u_int32_t	ar_parthreshold;	/* Thresh to determine peak */
	u_int32_t	ar_radarrssi;		/* Rssi threshold for AR event */
};

struct dfs_pulseparams {
	u_int64_t	p_time;			/* time for start of pulse in usecs*/
	u_int8_t	p_dur;			/* Duration of pulse in usecs*/
	u_int8_t	p_rssi;			/* Duration of pulse in usecs*/
};

struct dfs_pulseline {
	struct dfs_pulseparams pl_elems[DFS_MAX_PULSE_BUFFER_SIZE];	/* Array of pulses in delay line */
	u_int32_t	pl_firstelem;		/* Index of the first element */
	u_int32_t	pl_lastelem;		/* Index of the last element */
	u_int32_t	pl_numelems;		/* Number of elements in the delay line */
};

struct dfs_delayelem {
	u_int32_t	de_time;	/* Current "filter" time for start of pulse in usecs*/
	u_int32_t	de_dur;		/* Duration of pulse in usecs*/
	u_int32_t	de_rssi;	/* rssi of pulse in dB*/
};

/* NB: The first element in the circular buffer is the oldest element */

struct dfs_delayline {
	struct dfs_delayelem dl_elems[DFS_MAX_DL_SIZE];	/* Array of pulses in delay line */
	u_int64_t	dl_last_ts;		/* Last timestamp the delay line was used (in usecs) */
	u_int32_t	dl_firstelem;		/* Index of the first element */
	u_int32_t	dl_lastelem;		/* Index of the last element */
	u_int32_t	dl_numelems;		/* Number of elements in the delay line */
};

struct dfs_filter {
	struct dfs_delayline rf_dl;		/* Delay line of pulses for this filter */
	u_int32_t	rf_numpulses;		/* Number of pulses in the filter */
	u_int32_t	rf_pritype;		/* repitetion inter type, 1 -  fixed, 0 variable */
	u_int32_t 	rf_minpri;		/* min pri to be considered for this filter*/
	u_int32_t 	rf_maxpri;		/* max pri to be considered for this filter*/
	u_int32_t	rf_threshold;		/* match filter output threshold for radar detect */
	u_int32_t	rf_filterlen;		/* Length (in usecs) of the filter */
	u_int32_t	rf_mindur;		/* Min duration for this radar filter */
	u_int32_t	rf_maxdur;		/* Max duration for this radar filter */
	u_int32_t	rf_pulseid;		/* Unique ID corresponding to the original filter ID */
};

struct dfs_filtertype {
	struct dfs_filter ft_filters[DFS_MAX_NUM_RADAR_FILTERS];	/* array of filters */
	u_int32_t	ft_pritype;		/* repitetion inter type, 1 -  fixed, 0 variable */
	u_int32_t	ft_numfilters;		/* Num filters of this type */
	u_int64_t	ft_last_ts;		/* Last timestamp this filtertype was used (in usecs) */
	u_int32_t	ft_mindur;		/* min pulse duration to be considered
						   for this filter type */
	u_int32_t	ft_maxdur;		/* max pulse duration to be consdiered
						   for this filter type */
	u_int32_t	ft_rssithresh;		/* min rssi to be considered for this filter type */
	u_int32_t	ft_numpulses;		/* Num pulses in each filter of this type */
	u_int32_t 	ft_minpri;		/* min pri to be considered for this type*/
	u_int32_t	ft_rssimargin;		/* rssi threshold margin. In Turbo Mode HW reports rssi 3dB */
						/* lower than in non TURBO mode. This will offset that diff.*/
						/* lower than in non TURBO mode. This will offset that diff.*/
};

struct dfs_state {
	HAL_CHANNEL	rs_chan;		/* Channel info */
	u_int8_t	rs_chanindex;		/* Channel index in radar structure */
	u_int32_t	rs_numradarevents;	/* Number of radar events */
	int32_t		rs_firpwr;		/* Thresh to check radar sig is gone */ 
	u_int32_t	rs_radarrssi;		/* Thresh to start radar det (dB) */
	u_int32_t	rs_height;		/* Thresh for pulse height (dB)*/
	u_int32_t	rs_pulserssi;		/* Thresh to check if pulse is gone (dB) */
	u_int32_t	rs_inband;		/* Thresh to check if pulse is inband (0.5 dB) */
	/* 5413 specific */
	u_int32_t	rs_enrelpwr;		/* en/dis Relative power check */
	u_int32_t	rs_relpwr;		/* Relative power threshold in 0.5dB steps */
	u_int32_t	rs_enusefir128;		/* en/dis using 128 cycles */
	u_int32_t	rs_enblock;		/* en/dis block OFDM weak sig det as radar */
	u_int32_t	rs_enmaxrssi;		/* en/dis max rssi intead of last rssi */
	u_int32_t	rs_enrelstep;		/* en/dis relative step check */
	u_int32_t	rs_relstep;		/* Pulse Relative step threshold in 0.5dB steps */
	u_int32_t	rs_maxlen;		/* Max length of radar sign in 0.8us units */

};

struct dfs_nolelem {
	HAL_CHANNEL	nol_chan;		/* Channel info */
	u_int64_t	nol_tsffree;		/* 64 bit tsf when channel can be used again */
	struct dfs_nolelem *nol_next;		/* next element pointer */
};

struct dfs_info {
	int		rn_use_nol;		/* Use the NOL when radar found (default: TRUE) */
	u_int32_t	rn_numradars;		/* Number of different types of radars */
	u_int64_t	rn_lastfull_ts;		/* Last 64 bit timstamp from recv interrupt */
	u_int16_t	rn_last_ts;		/* last 15 bit ts from recv descriptor */
	u_int64_t	rn_ts_prefix;		/* Prefix to prepend to 15 bit recv ts */
	u_int32_t	rn_numbin5radars;	/* Number of bin5 radar pulses to search for */
	u_int32_t	rn_fastdivGCval;	/* Value of fast diversity gc limit from init file */
	int32_t		rn_minrssithresh;	/* Min rssi for all radar types */
	u_int32_t	rn_maxpulsedur;		/* Max pulse width in TSF ticks */
};

struct dfs_bin5pulses {
	u_int32_t	b5_threshold;		/* Number of bin5 pulses to indicate detection */
	u_int32_t	b5_mindur;		/* Min duration for a bin5 pulse */
	u_int32_t	b5_maxdur;		/* Max duration for a bin5 pulse */
	u_int32_t	b5_timewindow;		/* Window over which to count bin5 pulses */
	u_int32_t	b5_rssithresh;		/* Min rssi to be considered a pulse */
	u_int32_t	b5_rssimargin;  	/* rssi threshold margin. In Turbo Mode HW reports rssi 3dB */
                                        	/* lower than in non TURBO mode. This will be used to offset that diff.*/
};

struct dfs_bin5elem {
	u_int64_t	be_ts;			/* Timestamp for the bin5 element */
	u_int32_t	be_rssi;		/* Rssi for the bin5 element */
	u_int32_t	be_dur;			/* Duration of bin5 element */
};

struct dfs_bin5radars {
	struct dfs_bin5elem br_elems[DFS_MAX_B5_SIZE];	/* List of bin5 elems that fall
							 * within the time window */
	u_int32_t	br_firstelem;		/* Index of the first element */
	u_int32_t	br_lastelem;		/* Index of the last element */
	u_int32_t	br_numelems;		/* Number of elements in the delay line */
	struct dfs_bin5pulses br_pulse;		/* Original info about bin5 pulse */
};

struct dfs_stats {
        u_int32_t       num_radar_detects;      /* total num. of radar detects */
};

struct ath_dfs {
	int16_t			dfs_curchan_radindex;	/* cur. channel radar index */
	u_int32_t		dfsdomain;	/* cur. DFS domain */
	u_int32_t		dfs_proc_phyerr;/* Flags for Phy Errs to process */
	STAILQ_HEAD(,dfs_event)	dfs_eventq;	/* Q of free dfs event objects */
	dfsq_lock_t		dfs_eventqlock;	/* Lock for free dfs event list */
	unsigned int		dfs_eventqlockflags;	/* Flags for irq lock restore */
	STAILQ_HEAD(,dfs_event)	dfs_radarq;	/* Q of radar events */
	dfsq_lock_t		dfs_radarqlock;	/* Lock for dfs q */
	unsigned int		dfs_radarqlockflags;	/* Flags for irq lock restore */
	STAILQ_HEAD(,dfs_event) dfs_arq;	/* Q of AR events */
	dfsq_lock_t		dfs_arqlock;	/* Lock for AR q */
	unsigned int		dfs_arqlockflags;	/* Flags for irq lock restore */

	struct dfs_ar_state	dfs_ar_state;	/* AR state */
	struct dfs_state	dfs_radar[DFS_NUM_RADAR_STATES]; 	/* Per-Channel Radar detector state */
	struct dfs_filtertype	*dfs_radarf[DFS_MAX_RADAR_TYPES]; /* One filter for each radar pulse type */
	struct dfs_info		dfs_rinfo;	/* State variables for radar processing */
	struct dfs_bin5radars	*dfs_b5radars;	/* array of bin5 radar events */
	int8_t			**dfs_radartable;	/* mapping of radar durations to filter types */
	struct dfs_nolelem	*dfs_nol;	/* Non occupancy list for radar */
	HAL_PHYERR_PARAM	dfs_defaultparams; /* Default phy params per radar state */
        struct dfs_stats        stats;		/* DFS related stats */
	struct dfs_pulseline	pulses;		/* pulse history */
};

/* This should match the table from if_ath.c */
enum {
	ATH_DEBUG_DFS		= 0x00000100,	/* Minimal DFS debug */
	ATH_DEBUG_DFS1		= 0x00000200,	/* Normal DFS debug */
	ATH_DEBUG_DFS2		= 0x00000400,	/* Maximal DFS debug */
	ATH_DEBUG_DFS3		= 0x00000800,	/* matched filterID display */
};

/* These defines should match the table from ah_internal.h */
enum {
	DFS_UNINIT_DOMAIN	= 0,	/* Uninitialized dfs domain */
	DFS_FCC_DOMAIN		= 1,	/* FCC3 dfs domain */
	DFS_ETSI_DOMAIN		= 2,	/* ETSI dfs domain */
	DFS_MKK4_DOMAIN		= 3	/* Japan dfs domain */
};

#define	HAL_CAP_RADAR	0
#define	HAL_CAP_AR	1
#define HAL_CAP_STRONG_DIV 2

int		dfs_process_rdaarevent(struct ath_softc *sc, HAL_CHANNEL *chan);
u_int32_t	dfs_check_nol(struct ath_softc *sc, HAL_CHANNEL *chans,
			      u_int32_t nchans);
int		dfs_attach(struct ath_softc *sc);
void		dfs_detach(struct ath_softc *sc);
void		ath_ar_enable(struct ath_softc *sc);
int		dfs_radar_enable(struct ath_softc *sc);
void		ath_ar_disable(struct ath_softc *sc);
void		ath_process_phyerr(struct ath_softc *sc, struct ath_desc *ds,
				   u_int64_t fulltsf);
void		dfs_process_ar_event(struct ath_softc *sc, HAL_CHANNEL *chan);
struct dfs_state *dfs_getchanstate(struct ath_softc *sc, u_int8_t *index);
int		ath_ioctl_phyerr(struct ath_softc *sc, struct ath_diag *ad);


#endif  /* _DFS_H_ */
