/*
 * Copyright (c) 2002-2005 Atheros Communications, Inc.
 * All rights reserved.
 *
 * $Id: //depot/sw/releases/linuxsrc/src/802_11/madwifi/hal/main/ar5212/ar5416_eeprom.c#1 $
 */
#include "opt_ah.h"

#ifdef AH_SUPPORT_AR5416

#include "ah.h"
#include "ah_internal.h"
#include "ah_devid.h"
#include "ar5212/ar5416phy.h"
#include "ar5212/ar5416reg.h"
#include "ar5212/ar5416.h"

/*
 * Temporary cap on maximum power used in rate => power table
 * as workaround for board issues.
 */
#define TEMP_POWER_CAP      1

#define IS_5416_EMU_EEP(_ah)  ((AH_PRIVATE(_ah)->ah_devid == AR5416_DEVID_EMU_PCI) ||  (AH_PRIVATE(_ah)->ah_devid == AR5416_DEVID_EMU_PCIE))

/* reg_off = 4 * (eep_off) */
#define AR5416_EEPROM_S			2
#define AR5416_EEPROM_OFFSET		0x2000
#define AR5416_EEPROM_START_ADDR	0x503f1200
#define AR5416_EEPROM_MAX		0xae0 /* Ignore for the moment used only on the flash implementations */
#define owl_get_eep_ver(_ahp)   \
    (((_ahp)->baseEepHeader.version >> 12) & 0xF)
#define owl_get_eep_rev(_eep)   \
    (((_eep)->baseEepHeader.version) & 0xFFF)
#define owl_get_ntxchains(_txchainmask) \
    (((_txchainmask >> 2) & 1) + ((_txchainmask >> 1) & 1) + (_txchainmask & 1))

#define ah_5416eeprom			ah_eeprom.ee_u.eepromData5416

#ifdef __LINUX_ARM_ARCH__ /* AP71 */
#define owl_eep_start_loc		0
#define owl_eeprom_map(_ah)		ar5416FlashMap(_ah)
#define owl_eeprom_read			ar5416FlashRead
#define owl_eeprom_write		ar5416FlashWrite
#else
#define owl_eep_start_loc		256
#define owl_eeprom_map(_ah)
#define owl_eeprom_read			ar5416EepromRead
#define owl_eeprom_write		ar5416EepromWrite
#endif

//#define EEPROM_DUMP
/* End temp defines */

#define AR5416_OPFLAGS_11A           	1
#define AR5416_OPFLAGS_11G           	2

/* RF silent fields in EEPROM */
#define EEP_RFSILENT_ENABLED        	1
#define EEP_RFSILENT_POLARITY       	0x0002
#define EEP_RFSILENT_POLARITY_S     	1
#define EEP_RFSILENT_GPIO_SEL       	0x001c
#define EEP_RFSILENT_GPIO_SEL_S     	2

#define AR5416_EEP_NO_BACK_VER       	0x1
#define AR5416_EEP_VER               	0xE
// 16-bit offset location start of calibration struct
#define AR5416_EEP_START_LOC         	256
#define AR5416_NUM_5G_CAL_PIERS      	8
#define AR5416_NUM_2G_CAL_PIERS      	4
#define AR5416_NUM_5G_20_TARGET_POWERS  4
#define AR5416_NUM_5G_40_TARGET_POWERS  4
#define AR5416_NUM_2G_CCK_TARGET_POWERS 3
#define AR5416_NUM_2G_20_TARGET_POWERS  4
#define AR5416_NUM_2G_40_TARGET_POWERS  4
#define AR5416_NUM_CTLS              	16
#define AR5416_NUM_BAND_EDGES        	8
#define AR5416_NUM_PD_GAINS          	3
#define AR5416_PD_GAINS_IN_MASK      	4
#define AR5416_PD_GAIN_ICEPTS        	5
#define AR5416_EEPROM_MODAL_SPURS    	5
#define AR5416_MAX_RATE_POWER        	63
#define AR5416_NUM_PDADC_VALUES      	128
#define AR5416_NUM_RATES             	16
#define AR5416_BCHAN_UNUSED          	0xFF
#define AR5416_MAX_PWR_RANGE_IN_HALF_DB 64
#define AR5416_EEPMISC_BIG_ENDIAN    	0x01
#define FREQ2FBIN(x,y) 			((y) ? ((x) - 2300) : (((x) - 4800) / 5))
#define AR5416_MAX_CHAINS            	3
#define AR5416_ANT_16S               	25

#define AR5416_NUM_ANT_CHAIN_FIELDS     7
#define AR5416_NUM_ANT_COMMON_FIELDS    4
#define AR5416_SIZE_ANT_CHAIN_FIELD     3
#define AR5416_SIZE_ANT_COMMON_FIELD    4
#define AR5416_ANT_CHAIN_MASK           0x7
#define AR5416_ANT_COMMON_MASK          0xf
#define AR5416_CHAIN_0_IDX              0
#define AR5416_CHAIN_1_IDX              1
#define AR5416_CHAIN_2_IDX              2

typedef enum {
    EEP_NFTHRESH_5,
    EEP_NFTHRESH_2,
    EEP_MAC_MSW,
    EEP_MAC_MID,
    EEP_MAC_LSW,
    EEP_REG_0,
    EEP_REG_1,
    EEP_OP_CAP,
    EEP_OP_MODE,
    EEP_RF_SILENT,
} EEPROM_PARAM;

typedef enum Ar5416_Rates {
    rate6mb,  rate9mb,  rate12mb, rate18mb,
    rate24mb, rate36mb, rate48mb, rate54mb,
    rate1l,   rate2l,   rate2s,   rate5_5l,
    rate5_5s, rate11l,  rate11s,  rateXr,
    rateHt20_0, rateHt20_1, rateHt20_2, rateHt20_3,
    rateHt20_4, rateHt20_5, rateHt20_6, rateHt20_7,
    rateHt40_0, rateHt40_1, rateHt40_2, rateHt40_3,
    rateHt40_4, rateHt40_5, rateHt40_6, rateHt40_7,
    rateDupCck, rateDupOfdm, rateExtCck, rateExtOfdm,
    Ar5416RateSize
} AR5416_RATES;

typedef struct BaseEepHeader {
    u_int16_t  length;
    u_int16_t  checksum;
    u_int16_t  version;
    u_int8_t   opCapFlags;
    u_int8_t   eepMisc;
    u_int16_t  regDmn[2];
    u_int8_t   macAddr[6];
    u_int8_t   rxMask;
    u_int8_t   txMask;
    u_int16_t  rfSilent;
    u_int16_t  blueToothOptions;
    u_int16_t  deviceCap;
    u_int8_t   futureBase[38];
} __packed BASE_EEP_HEADER; // 64 B

typedef struct spurChanStruct {
    u_int16_t spurChan;
    u_int8_t  spurRangeLow;
    u_int8_t  spurRangeHigh;
} __packed SPUR_CHAN;

typedef struct ModalEepHeader {
    u_int32_t  antCtrlChain[AR5416_MAX_CHAINS];       // 12
    u_int32_t  antCtrlCommon;                         // 4
    u_int8_t   antennaGainCh[AR5416_MAX_CHAINS];      // 3
    u_int8_t   switchSettling;                        // 1
    u_int8_t   txRxAttenCh[AR5416_MAX_CHAINS];        // 3
    u_int8_t   rxTxMarginCh[AR5416_MAX_CHAINS];       // 3
    u_int8_t   adcDesiredSize;                        // 1
    u_int8_t   pgaDesiredSize;                        // 1
    u_int8_t   xlnaGainCh[AR5416_MAX_CHAINS];         // 3
    u_int8_t   txEndToXpaOff;                         // 1
    u_int8_t   txEndToRxOn;                           // 1
    u_int8_t   txFrameToXpaOn;                        // 1
    u_int8_t   thresh62;                              // 1
    u_int8_t   noiseFloorThreshCh[AR5416_MAX_CHAINS]; // 3
    u_int8_t   xpdGain;                               // 1
    u_int8_t   xpd;                                   // 1
    u_int8_t   iqCalICh[AR5416_MAX_CHAINS];           // 1
    u_int8_t   iqCalQCh[AR5416_MAX_CHAINS];           // 1
    u_int8_t   pdGainOverlap;                         // 1
    u_int8_t   ob;                                    // 1
    u_int8_t   db;                                    // 1
    u_int8_t   xpaBiasLvl;                            // 1
    u_int8_t   pwrDecreaseFor2Chain;                  // 1
    u_int8_t   pwrDecreaseFor3Chain;                  // 1 -> 48 B
    u_int8_t   futureModal[32];                       // 32 B
    SPUR_CHAN spurChans[AR5416_EEPROM_MODAL_SPURS];  // 20 B
} __packed MODAL_EEP_HEADER;                    // == 100 B

typedef struct calDataPerFreq {
    u_int8_t pwrPdg[AR5416_NUM_PD_GAINS+1][AR5416_PD_GAIN_ICEPTS];
    u_int8_t vpdPdg[AR5416_NUM_PD_GAINS+1][AR5416_PD_GAIN_ICEPTS];
} __packed CAL_DATA_PER_FREQ;

typedef struct CalTargetPowerLegacy {
    u_int8_t  bChannel;
    u_int8_t  tPow2x[4];
} __packed CAL_TARGET_POWER_LEG;

typedef struct CalTargetPowerHt {
    u_int8_t  bChannel;
    u_int8_t  tPow2x[8];
} __packed CAL_TARGET_POWER_HT;

#ifdef BIG_ENDIAN
typedef struct CalCtlEdges {
    u_int8_t  bChannel;
    u_int8_t  flag   :2,
             tPower :6;
} __packed CAL_CTL_EDGES;
#else
typedef struct CalCtlEdges {
    u_int8_t  bChannel;
    u_int8_t  tPower :6,
             flag   :2;
} __packed CAL_CTL_EDGES;
#endif

typedef struct CalCtlData {
    CAL_CTL_EDGES  ctlEdges[AR5416_MAX_CHAINS][AR5416_NUM_BAND_EDGES];
} __packed CAL_CTL_DATA;

typedef struct eepMap {
    BASE_EEP_HEADER      baseEepHeader;         // 64 B
    u_int8_t             custData[64];          // 64 B
    MODAL_EEP_HEADER     modalHeader[2];        // 200 B
    u_int8_t             calFreqPier5G[AR5416_NUM_5G_CAL_PIERS];
    u_int8_t             calFreqPier2G[AR5416_NUM_2G_CAL_PIERS];
    CAL_DATA_PER_FREQ    calPierData5G[AR5416_MAX_CHAINS][AR5416_NUM_5G_CAL_PIERS];
    CAL_DATA_PER_FREQ    calPierData2G[AR5416_MAX_CHAINS][AR5416_NUM_2G_CAL_PIERS];
    CAL_TARGET_POWER_LEG calTargetPower5G[AR5416_NUM_5G_20_TARGET_POWERS];
    CAL_TARGET_POWER_HT  calTargetPower5GHT20[AR5416_NUM_5G_20_TARGET_POWERS];
    CAL_TARGET_POWER_HT  calTargetPower5GHT40[AR5416_NUM_5G_40_TARGET_POWERS];
    CAL_TARGET_POWER_LEG calTargetPowerCck[AR5416_NUM_2G_CCK_TARGET_POWERS];
    CAL_TARGET_POWER_LEG calTargetPower2G[AR5416_NUM_2G_20_TARGET_POWERS];
    CAL_TARGET_POWER_HT  calTargetPower2GHT20[AR5416_NUM_2G_20_TARGET_POWERS];
    CAL_TARGET_POWER_HT  calTargetPower2GHT40[AR5416_NUM_2G_40_TARGET_POWERS];
    u_int8_t             ctlIndex[AR5416_NUM_CTLS];
    CAL_CTL_DATA         ctlData[AR5416_NUM_CTLS];
} __packed ar5416_eeprom_t;

/* Fake Owl Eeprom Yeech...*/
#define OWLEMU_EEPROM_SZ        1024
static u_int16_t owlemu_eeprom[OWLEMU_EEPROM_SZ] = {
0x0013, 0x168c, 0x0200, 0x0001, 0x0000, 0x5001, 0x0000, 0x1042,
0x168c, 0x1c0a, 0x0100, 0x0000, 0x01c2, 0x0002, 0xc606, 0x0001,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x43b5, 0x7f03, 0x0003,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x5aa5, 0x0000, 0x0000,
0x0313, 0x4943, 0x2053, 0x7104, 0x1202, 0x0400, 0x0306, 0x0001,
0x0000, 0x0500, 0x410e, 0x39b1, 0x1eb5, 0x4e2d, 0x3056, 0xffff,
0xe902, 0x0700, 0x0106, 0x0000, 0x0100, 0x1500, 0x0752, 0x4101,
0x6874, 0x7265, 0x736f, 0x4320, 0x6d6f, 0x756d, 0x696e, 0x6163,
0x6974, 0x6e6f, 0x2c73, 0x4920, 0x636e, 0x002e, 0x5241, 0x3035,
0x3130, 0x302d, 0x3030, 0x2d30, 0x3030, 0x3030, 0x5700, 0x7269,
0x6c65, 0x7365, 0x2073, 0x414c, 0x204e, 0x6552, 0x6566, 0x6572,
0x636e, 0x2065, 0x6143, 0x6472, 0x3000, 0x0030, 0x00ff, 0x2100,
0x0602, 0x2201, 0x0205, 0x8d80, 0x005b, 0x0522, 0x4002, 0x8954,
0x2200, 0x0205, 0x1b00, 0x00b7, 0x0522, 0x8002, 0x12a8, 0x2201,
0x0205, 0x3600, 0x016e, 0x0522, 0x0002, 0x2551, 0x2202, 0x0205,
0x6c00, 0x02dc, 0x0522, 0x8002, 0x37f9, 0x2203, 0x0205, 0xa200,
0x044a, 0x0222, 0x0803, 0x0822, 0x0604, 0x0300, 0x037f, 0xb543,
0x0222, 0x0105, 0x00ff, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x4243, 0x3234, 0x3035, 0x5f4c, 0x3930, 0x3934, 0x3061, 0xffff,
0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xf000, 0x2094, 0x0000,
0xe839, 0x4008, 0x0a07, 0x0000, 0x420a, 0x41a0, 0x0101, 0x3002,
0x0003, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x2d3c, 0x0081, 0x0000, 0x0108,
0x0000, 0xe049, 0x2492, 0x020f, 0x000e, 0xb0ca, 0x21a3, 0x4024,
0x08f9, 0x000b, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x2398, 0x0091, 0x0410, 0x4148, 0x2082, 0xda22,
0x021c, 0x0007, 0xb0ff, 0x01a3, 0x4012, 0x0001, 0xa270, 0x17b8,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x3198, 0x0091, 0x0410,
0x4148, 0x2082, 0xe022, 0x021c, 0x000e, 0xb0ff, 0x21a3, 0x4012,
0x7851, 0xa270, 0x1320, 0x17b8, 0x0080, 0x0003, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x1013, 0x1112, 0x1440, 0x3031, 0x3234, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x3214, 0x5c4a, 0x7868, 0xb9a0, 0x00cd, 0x3619, 0x5543, 0x1d55,
0x0000, 0x1900, 0x381e, 0x5346, 0x1534, 0x0000, 0x1900, 0x3e1f,
0x5252, 0x0935, 0x0000, 0x1900, 0x3e23, 0x5252, 0x0534, 0x0000,
0x1900, 0x3e25, 0x5252, 0x0533, 0x0000, 0x1900, 0x3e25, 0x5252,
0x0532, 0x0000, 0x1900, 0x412b, 0x5353, 0x0510, 0x0000, 0x1900,
0x4430, 0x5750, 0x08ef, 0x0000, 0x1900, 0x4334, 0x574b, 0x10cc,
0x0000, 0x1900, 0x4a2f, 0x5858, 0x00f4, 0x0000, 0x1900, 0x4c30,
0x5858, 0x00f4, 0x0000, 0x1900, 0x4429, 0x5854, 0x0933, 0x0000,
0x1900, 0x4a2d, 0x5c56, 0x0534, 0x0000, 0x1900, 0x482d, 0x5c57,
0x0534, 0x0000, 0x1900, 0x4a2d, 0x5c58, 0x0535, 0x0000, 0x1900,
0x4c91, 0xe71a, 0x5089, 0xe71a, 0x6889, 0xe71a, 0x8c89, 0xe71a,
0xb489, 0xe71a, 0xbd89, 0xe71a, 0xcd89, 0xe71a, 0xd289, 0xe71a,
0x70a2, 0x8a28, 0xb8a2, 0x8a28, 0x709a, 0x481e, 0x939a, 0x481e,
0xac9a, 0x481e, 0x4c58, 0x5c68, 0xbdc1, 0xcd00, 0x2222, 0x2423,
0x2464, 0x2400, 0x5052, 0x5a60, 0x62c0, 0xc1c9, 0x2262, 0x2224,
0x2423, 0x6323, 0x7075, 0xa200, 0x0000, 0x0000, 0x2868, 0x2800,
0x0000, 0x0000, 0x7075, 0x9da2, 0x0000, 0x0000, 0x2166, 0x6621,
0x0000, 0x0000, 0x8989, 0x0000, 0x0000, 0x0000, 0x2626, 0x0000,
0x0000, 0x0000, 0x4a56, 0x0000, 0x0000, 0x0000, 0x3c3c, 0x0000,
0x0000, 0x0000, 0x4c68, 0x8cb4, 0xbdc1, 0xcd00, 0x2424, 0x2424,
0x3c7c, 0x3c00, 0x7075, 0xac00, 0x0000, 0x0000, 0x2464, 0x2300,
0x0000, 0x0000, 0x7075, 0xac00, 0x0000, 0x0000, 0x2262, 0x2100,
0x0000, 0x0000, 0x8989, 0x0000, 0x0000, 0x0000, 0x2222, 0x0000,
0x0000, 0x0000, 0xea0e, 0xc000, 0x0e07, 0x8042, 0xa204, 0x0002,
0x0e03, 0x8186, 0x9858, 0x0000, 0x0e01, 0x800c, 0x9824, 0x0707,
0x0e02, 0x8063, 0x9860, 0x0003, 0x03ff, 0xa210, 0x0080, 0x6333,
0x03ff, 0xa214, 0x0010, 0x6c10, 0x03ff, 0xa21c, 0x1883, 0x800a,
0x03ff, 0xa220, 0x0188, 0x30c6, 0x05ff, 0xc0fd, 0x1001, 0xc8f5,
0x707f, 0xc8fc, 0x1000, 0xc85a, 0x2001, 0xc85c, 0x2002, 0xc85e,
0x2001, 0xc8fe, 0x1001, 0xc519, 0x1001, 0xcc01, 0x2001, 0xcc03,
0x2001, 0xcc8b, 0x1001, 0xcc8c, 0x1001, 0xdc92, 0x1001, 0x0fff,
0x8081, 0x1230, 0x0001, 0x0fff, 0x8010, 0x982c, 0xa002, 0x09ff,
0x8063, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
};

static void ar5416FillEmuEeprom(struct ath_hal *ahp);
static HAL_BOOL ar5416FillEeprom(struct ath_hal *ah);
static HAL_STATUS ar5416CheckEeprom(struct ath_hal *ah);
static HAL_BOOL ar5416SetPowerPerRateTable(struct ath_hal *ah,
                           ar5416_eeprom_t *pEepData, HAL_HT *ht,
                           HAL_CHANNEL_INTERNAL *chan, int16_t *ratesArray,
                           u_int16_t cfgCtl, u_int16_t AntennaReduction,
                           u_int16_t twiceMaxRegulatoryPower, 
                           u_int16_t powerLimit);
static HAL_BOOL ar5416SetPowerCalTable(struct ath_hal *ah,
			HAL_HT *ht, ar5416_eeprom_t *pEepData,
			HAL_CHANNEL_INTERNAL *chan,
			int16_t *pTxPowerIndexOffset);
static u_int16_t ar5416GetMaxEdgePower(u_int16_t freq,
			CAL_CTL_EDGES *pRdEdgesPower, HAL_BOOL is2GHz);
static void ar5416GetTargetPowers(struct ath_hal *ah, HAL_HT *ht,
			HAL_CHANNEL_INTERNAL *chan, CAL_TARGET_POWER_HT *powInfo,
			u_int16_t numChannels, CAL_TARGET_POWER_HT *pNewPower,
			u_int16_t numRates, HAL_BOOL isHt40Target);
static void ar5416GetTargetPowersLeg(struct ath_hal *ah, HAL_HT *ht,
			HAL_CHANNEL_INTERNAL *chan, CAL_TARGET_POWER_LEG *powInfo,
			u_int16_t numChannels, CAL_TARGET_POWER_LEG *pNewPower,
			u_int16_t numRates, HAL_BOOL isExtTarget);

static u_int16_t fbin2freq(u_int8_t fbin, HAL_BOOL is2GHz);
static int16_t interpolate(u_int16_t target, u_int16_t srcLeft,
			u_int16_t srcRight, int16_t targetLeft,
			int16_t targetRight);
static void ar5416GetGainBoundariesAndPdadcs(struct ath_hal *ah, HAL_HT *ht,
                        HAL_CHANNEL_INTERNAL *chan, CAL_DATA_PER_FREQ *pRawDataSet,
                        u_int8_t * bChans, u_int16_t availPiers,
                        u_int16_t tPdGainOverlap, int16_t *pMinCalPower,
                        u_int16_t * pPdGainBoundaries, u_int8_t * pPDADCValues,
                        u_int16_t numXpdGains);
static HAL_BOOL getLowerUpperIndex(u_int8_t target, u_int8_t *pList,
			u_int16_t listSize,  u_int16_t *indexL,
 			u_int16_t *indexR);
static HAL_BOOL ar5416FillVpdTable(u_int8_t pwrMin, u_int8_t pwrMax,
			u_int8_t *pPwrList, u_int8_t *pVpdList,
			u_int16_t numIntercepts, u_int8_t *pRetVpdList);
			  
static u_int16_t ar5416EepromGetSpurChan(struct ath_hal *ah, u_int16_t spurChan,HAL_BOOL is2GHz);
static HAL_BOOL ar5416GetChipPowerLimits(struct ath_hal *ah, HAL_CHANNEL *chans, u_int32_t nchans);
static HAL_BOOL ar5416GetEepromCapabilityInfo(struct ath_hal *ah, u_int16_t *capField);
static void ar5416ReadEepromCTLInfo(struct ath_hal *ah);

/* Functions exported via attach */
//HAL_STATUS ar5416EepromAttach(struct ath_hal *ah);
static HAL_STATUS ar5416EepromDetach(struct ath_hal *ah);
static u_int32_t ar5416EepromGet(struct ath_hal *ahp, EEPROM_PARAM param);
static HAL_BOOL ar5416EepromSetBoardValues(struct ath_hal *ah, HAL_CHANNEL_INTERNAL *chan);
static HAL_BOOL ar5416SetTransmitPower(struct ath_hal *ah, 
	HAL_CHANNEL_INTERNAL *chan, u_int16_t *rfXpdGain);
static HAL_BOOL ar5416GetEepromNoiseFloorThresh(struct ath_hal *ah, 
			const HAL_CHANNEL_INTERNAL *chan, int16_t *nft);
#ifdef EEPROM_DUMP	
static void ar5416PrintPowerPerRate(struct ath_hal *ah, int16_t pRatesPower[]);
static void ar5416EepromDump(struct ath_hal *ah, ar5416_eeprom_t *ar5416Eep);
#endif
/*****************************
 * Eeprom APIs for CB/XB/MB only
 ****************************/

/*
 * Read 16 bits of data from offset into *data
 */
HAL_BOOL
ar5416EepromRead(struct ath_hal *ah, u_int off, u_int16_t *data)
{

	if (IS_5416_EMU_EEP(ah)) {
        	off = off & (OWLEMU_EEPROM_SZ - 1);
        	*data = owlemu_eeprom[off];
		return AH_TRUE;
	} else {

        	(void)OS_REG_READ(ah,  AR5416_EEPROM_OFFSET + (off << AR5416_EEPROM_S) );
       		if (!ath_hal_wait(ah, AR_EEPROM_STATUS_DATA, AR_EEPROM_STATUS_DATA_BUSY
				| AR_EEPROM_STATUS_DATA_PROT_ACCESS, 0)) {
            	return AH_FALSE;
        	}

       	 	*data = MS(OS_REG_READ(ah, AR_EEPROM_STATUS_DATA), AR_EEPROM_STATUS_DATA_VAL);
        	return AH_TRUE;
	}

}

#ifdef AH_SUPPORT_WRITE_EEPROM
/*
 * Write 16 bits of data from data to the specified EEPROM offset.
 */
HAL_BOOL
ar5416EepromWrite(struct ath_hal *ah, u_int off, u_int16_t data)
{
	if (!IS_5416_EMU_EEP(ah))
		OS_REG_WR(ah, AR5416_EEPROM_OFFSET + (off << AR5416_EEPROM_S), data);
        return AH_TRUE;
}
#endif /* AH_SUPPORT_WRITE_EEPROM */

/*************************
 * Flash APIs for AP only
 *************************/
#if 0
HAL_STATUS
ar5416FlashMap(struct ath_hal *ah)
{
    struct ath_hal *ahp = AH5416(ah);

    ahp->ah_cal_mem = OS_REMAP(AR5416_EEPROM_START_ADDR, AR5416_EEPROM_MAX);
    if (!ahp->ah_cal_mem) {
        HALDEBUG(ah, "%s: cannot remap eeprom region \n", __func__);
        return HAL_EIO;
    }

    return HAL_OK;
}

HAL_BOOL
ar5416FlashRead(struct ath_hal *ah, u_int off, u_int16_t *data)
{
    struct ath_hal *ahp = AH5416(ah);

    *data = ((u_int16_t *)ahp->ah_cal_mem)[off];
    return AH_TRUE;
}

HAL_BOOL
ar5416FlashWrite(struct ath_hal *ah, u_int off, u_int16_t data)
{
    struct ath_hal *ahp = AH5416(ah);

    ((u_int16_t *)ahp->ah_cal_mem)[off] = data;
    return AH_TRUE;
}
#endif
/***************************
 * Common APIs for AP/CB/XB
 ***************************/

HAL_STATUS
ar5416EepromAttach(struct ath_hal *ah)
{
    struct ath_hal_5416 *ahp = AH5416(ah);
    HAL_STATUS status;

    AH_PRIVATE(ah)->ah_eepromRead = ar5416EepromRead;
    AH_PRIVATE(ah)->ah_eepromDetach = ar5416EepromDetach;
    AH_PRIVATE(ah)->ah_eepromSetBoardValues = ar5416EepromSetBoardValues;
    AH_PRIVATE(ah)->ah_eepromGetSpurChan =  ar5416EepromGetSpurChan;
    AH_PRIVATE(ah)->ah_setTransmitPower = ar5416SetTransmitPower;
    AH_PRIVATE(ah)->ah_getChipPowerLimits = ar5416GetChipPowerLimits;
    AH_PRIVATE(ah)->ah_getEepromCapabilityInfo = ar5416GetEepromCapabilityInfo;
    AH_PRIVATE(ah)->ah_getEepromNoiseFloorThresh = ar5416GetEepromNoiseFloorThresh;

    //owl_eeprom_map(ah);
    /* Allocate memory for copy of eeprom */
    ahp->ah_5416eeprom = ath_hal_malloc(sizeof(ar5416_eeprom_t));
    if (ahp->ah_5416eeprom == AH_NULL) {
    	HALDEBUG(ah, "%s: Failed memory allocation for eeprom\n", __func__);
    	return HAL_ENOMEM;
    }

    ath_hal_memzero(ahp->ah_5416eeprom,sizeof(ar5416_eeprom_t));

    if (IS_5416_EMU_EEP(ah)) { 
    	ar5416FillEmuEeprom((struct ath_hal *)ahp);
	return HAL_OK;
    }

    if (!ar5416FillEeprom(ah)){
    	HALDEBUG(ah,"%s Error reading AR5416/AR5418 Eeprom.\n",__func__);
        return HAL_EIO;
    }

    if ((status = ar5416CheckEeprom(ah)) != HAL_OK) {
    	HALDEBUG(ah,"%s AR5416/AR5418 Eeprom failed check.\n",__func__);
    	return status;
    }
     /* Get CTLs */
    ar5416ReadEepromCTLInfo(ah);
    
    HALDEBUG(ah,"%s ahp->ah_5416eeprom=0x%p\n",__func__, ahp->ah_5416eeprom);

    return HAL_OK;
}
static HAL_STATUS
ar5416EepromDetach(struct ath_hal *ah) {
	struct ath_hal_5416 *ahp = AH5416(ah);
	
	if (ahp->ah_5416eeprom)
		ath_hal_free(ahp->ah_5416eeprom);
		
	return HAL_OK;
}
static u_int32_t
ar5416EepromGet(struct ath_hal *ahp, EEPROM_PARAM param)
{
    ar5416_eeprom_t *eep = (ar5416_eeprom_t *)(AH5416(ahp)->ah_5416eeprom);
    MODAL_EEP_HEADER *pModal = eep->modalHeader;
    BASE_EEP_HEADER  *pBase  = &eep->baseEepHeader;

    switch (param) {
        case EEP_NFTHRESH_5:
    	 /* 5GHz Threshold is a signed value.
	  * Use cast to ensure proper sign extension
	  */	
            return (int8_t)pModal[0].noiseFloorThreshCh[0];
        case EEP_NFTHRESH_2:
	 /* 2GHz Threshold is a signed value. 
	  * Use cast to ensure proper sign extension
	  */
            return (int8_t)pModal[1].noiseFloorThreshCh[0];
        case AR_EEPROM_MAC(0):
            return pBase->macAddr[0] << 8 | pBase->macAddr[1];
        case AR_EEPROM_MAC(1):
            return pBase->macAddr[2] << 8 | pBase->macAddr[3];
        case AR_EEPROM_MAC(2):
            return pBase->macAddr[4] << 8 | pBase->macAddr[5];
        case EEP_REG_0:
            return pBase->regDmn[0];
        case EEP_REG_1:
            return pBase->regDmn[1];
        case EEP_OP_CAP:
            return pBase->deviceCap;
        case EEP_OP_MODE:
            return pBase->opCapFlags;
        case EEP_RF_SILENT:
            return pBase->rfSilent;
        default:
            HALASSERT(0);
            return 0;
    }
}

/*
 * Read EEPROM header info and program the device for correct operation
 * given the channel value.
 */
static HAL_BOOL
ar5416EepromSetBoardValues(struct ath_hal *ah, HAL_CHANNEL_INTERNAL *chan)
{
    MODAL_EEP_HEADER *pModal;
    int i, regChainOffset;
    struct ath_hal_5212 *ahp = AH5416(ah);
    ar5416_eeprom_t *eep = (ar5416_eeprom_t *)ahp->ah_5416eeprom;

    HALASSERT(owl_get_eep_ver(ahp) == AR5416_EEP_VER);
    pModal = &(eep->modalHeader[IS_CHAN_2GHZ(chan)]);

    if (IS_5416_EMU_EEP(ah))
        return AH_TRUE;

    OS_REG_WRITE(ah, AR_PHY_SWITCH_COM, pModal->antCtrlCommon);
    for (i = 0; i < AR5416_MAX_CHAINS; i++) {
       // if((AH_PRIVATE(ah)->ah_macRev >= AR_SREV_REVISION_OWL_20) && 
       	   if (!(IS_5416V1(ah)) &&
           (eep->baseEepHeader.txMask == 0x5) && (i != 0)) {
            /* Regs are swapped from chain 2 to 1 for 5416 2_0 with 
             * only chains 0 and 2 populated 
             */
            regChainOffset = (i == 1) ? 0x2000 : 0x1000;
        } else {
            regChainOffset = i * 0x1000;
        }

        OS_REG_WRITE(ah, AR_PHY_SWITCH_CHAIN_0 + regChainOffset, pModal->antCtrlChain[i]);
        OS_REG_WRITE(ah, AR_PHY_TIMING_CTRL4 + regChainOffset, 
        (OS_REG_READ(ah, AR_PHY_TIMING_CTRL4 + regChainOffset) &
        ~(AR_PHY_TIMING_CTRL4_IQCORR_Q_Q_COFF | AR_PHY_TIMING_CTRL4_IQCORR_Q_I_COFF)) |
        SM(pModal->iqCalICh[i], AR_PHY_TIMING_CTRL4_IQCORR_Q_I_COFF) |
        SM(pModal->iqCalQCh[i], AR_PHY_TIMING_CTRL4_IQCORR_Q_Q_COFF));

        if ((i == 0) || !IS_5416V1(ah)) {
            //(AH_PRIVATE(ah)->ah_macRev >= AR_SREV_REVISION_OWL_20)) {
            OS_REG_WRITE(ah, AR_PHY_RXGAIN + regChainOffset, 
            (OS_REG_READ(ah, AR_PHY_RXGAIN + regChainOffset) & ~AR_PHY_RXGAIN_TXRX_ATTEN) |
            SM(pModal->txRxAttenCh[i], AR_PHY_RXGAIN_TXRX_ATTEN));
            OS_REG_WRITE(ah, AR_PHY_GAIN_2GHZ + regChainOffset, (OS_REG_READ(ah, AR_PHY_GAIN_2GHZ + regChainOffset) & ~AR_PHY_GAIN_2GHZ_RXTX_MARGIN) |
                SM(pModal->rxTxMarginCh[i], AR_PHY_GAIN_2GHZ_RXTX_MARGIN));
        }
    }
    /* write previous IQ results */
    OS_REG_SET_BIT(ah, AR_PHY_TIMING_CTRL4, AR_PHY_TIMING_CTRL4_IQCORR_ENABLE);

    OS_REG_RMW_FIELD(ah, AR_PHY_SETTLING, AR_PHY_SETTLING_SWITCH, pModal->switchSettling);
    OS_REG_RMW_FIELD(ah, AR_PHY_DESIRED_SZ, AR_PHY_DESIRED_SZ_ADC, pModal->adcDesiredSize);
    OS_REG_RMW_FIELD(ah, AR_PHY_DESIRED_SZ, AR_PHY_DESIRED_SZ_PGA, pModal->pgaDesiredSize);
    OS_REG_WRITE(ah, AR_PHY_RF_CTL4,
        SM(pModal->txEndToXpaOff, AR_PHY_RF_CTL4_TX_END_XPAA_OFF)
        | SM(pModal->txEndToXpaOff, AR_PHY_RF_CTL4_TX_END_XPAB_OFF)
        | SM(pModal->txFrameToXpaOn, AR_PHY_RF_CTL4_FRAME_XPAA_ON)
        | SM(pModal->txFrameToXpaOn, AR_PHY_RF_CTL4_FRAME_XPAB_ON));

    OS_REG_RMW_FIELD(ah, AR_PHY_RF_CTL3, AR_PHY_TX_END_TO_A2_RX_ON, pModal->txEndToRxOn);
    OS_REG_RMW_FIELD(ah, AR_PHY_CCA, AR_PHY_CCA_THRESH62, pModal->thresh62);

    return AH_TRUE;
}
#if 0
/**************************************************************
 * ar5416GetEChan
 *
 * Construct EEPROM channel makeup from local structs
 */
void
ar5416GetEChan(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval, A_UCHAR txChainSelect, HAL_HT *ht, EEPROM_CHANNEL *pEChan)
{
    ASSERT(txChainSelect); /* Otherwise negative indexing in EEPROM */
    pEChan->activeTxChains = ( ((txChainSelect >> 2) & 1) +
        ((txChainSelect >> 1) & 1) + (txChainSelect & 1));
    pEChan->isHt40 = IS_CHAN_HT40(pChval->channelFlags) && (ht->ht_phymode == HAL_HT_PHYMODE_2040);
    pEChan->is2GHz = IS_CHAN_2GHZ(pChval->channelFlags);
    if (pEChan->isHt40 && ht) {
        int ctlDir;
        pEChan->synthCenter = pEChan->ht40Center = ar5416GetChannelCenter(pDev, ht, pChval);
        ctlDir = ((ht->ht_extoff == 1) ? (1) : (-1));
        pEChan->ctlCenter = (A_UINT16)(pEChan->synthCenter + (ctlDir * 10));
        pEChan->extCenter = (A_UINT16)(pEChan->synthCenter - (ctlDir * ((ht->ht_extprotspacing == HAL_HT_EXTPROTSPACING_20) ? (10) : (15))));
    } else {
        pEChan->synthCenter = pEChan->ht40Center = pEChan->ctlCenter = pEChan->extCenter = pChval->channel;
    }
}
#endif

/**************************************************************
 * ar5416SetTransmitPower
 *
 * Set the transmit power in the baseband for the given
 * operating channel and mode.
 */
static HAL_BOOL
ar5416SetTransmitPower(struct ath_hal *ah, HAL_CHANNEL_INTERNAL *chan, u_int16_t *rfXpdGain)
{
#define POW_SM(_r, _s)     (((_r) & 0x3f) << (_s))
#define N(a)            (sizeof (a) / sizeof (a[0]))

    MODAL_EEP_HEADER *pModal;
    struct ath_hal_5416 *ahp;
    //int16_t ratesArray[Ar5416RateSize] = { 0 };
    int16_t ratesArray[Ar5416RateSize];
    int16_t  txPowerIndexOffset = 0;
    int i;
    HAL_HT	*ht;
    
    u_int16_t cfgCtl;
    u_int16_t powerLimit;
    u_int16_t twiceAntennaReduction;
    u_int16_t twiceMaxRegulatoryPower;
    ar5416_eeprom_t *pEepData; 
    ahp = AH5416(ah);
    
    HALASSERT(owl_get_eep_ver(ahp) == AR5416_EEP_VER);

    if (IS_5416_EMU_EEP(ah))
         return AH_TRUE;

    /* Setup info for the actual eeprom */
    ath_hal_memzero(ratesArray,sizeof(ratesArray));
    cfgCtl = ath_hal_getctl(ah, (HAL_CHANNEL *)chan);
    powerLimit = chan->maxRegTxPower * 2;
    twiceAntennaReduction = chan->antennaMax;
    twiceMaxRegulatoryPower = AH_MIN(MAX_RATE_POWER, AH_PRIVATE(ah)->ah_powerLimit); 
    pEepData = (ar5416_eeprom_t *)ahp->ah_5416eeprom;
    ht = ahp->ah_ht;
    pModal = &(pEepData->modalHeader[IS_CHAN_2GHZ(chan)]);
    HALDEBUG(ah,"%s Channel=%u CfgCtl=%u\n",
    			__func__,chan->channel, cfgCtl );      
  
    if (!ar5416SetPowerPerRateTable(ah, pEepData, ht, chan,
                                    &ratesArray[0],cfgCtl,
                                    twiceAntennaReduction,
				    twiceMaxRegulatoryPower, powerLimit)) {
        HALDEBUG(ah, "ar5416EepromSetTransmitPower: unable to set tx power per rate table\n");
        return AH_FALSE;
    }

    if (!ar5416SetPowerCalTable(ah, ht, pEepData, chan, &txPowerIndexOffset)) {
        HALDEBUG(ah, "ar5416EepromSetTransmitPower: unable to set power table\n");
        return AH_FALSE;
    }
    /*
     * txPowerIndexOffset is set by the SetPowerTable() call -
     *  adjust the rate table (0 offset if rates EEPROM not loaded)
     */
    for (i = 0; i < N(ratesArray); i++) {
        ratesArray[i] = (int16_t)(txPowerIndexOffset + ratesArray[i]);
        if (ratesArray[i] > AR5416_MAX_RATE_POWER)
            ratesArray[i] = AR5416_MAX_RATE_POWER;
#ifdef TEMP_POWER_CAP
        /*
         * workaround: cap on max power 
         * ratesArray table is in half dB steps
         */
        if (ratesArray[i] > 2 * ath_hal_maxTPC) {
            ratesArray[i] = 2 * ath_hal_maxTPC;
        }
#endif
    }

#ifdef EEPROM_DUMP
    ar5416PrintPowerPerRate(ah, ratesArray);
#endif

    /* Write the OFDM power per rate set */
    OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE1,
        POW_SM(ratesArray[rate18mb], 24)
          | POW_SM(ratesArray[rate12mb], 16)
          | POW_SM(ratesArray[rate9mb],  8)
          | POW_SM(ratesArray[rate6mb],  0)
    );
    OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE2,
        POW_SM(ratesArray[rate54mb], 24)
          | POW_SM(ratesArray[rate48mb], 16)
          | POW_SM(ratesArray[rate36mb],  8)
          | POW_SM(ratesArray[rate24mb],  0)
    );

    if (IS_CHAN_2GHZ(chan)) {
        /* Write the CCK power per rate set */
        OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE3,
            POW_SM(ratesArray[rate2s], 24)
              | POW_SM(ratesArray[rate2l],  16)
              | POW_SM(ratesArray[rateXr],  8) /* XR target power */
              | POW_SM(ratesArray[rate1l],   0)
        );
        OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE4,
            POW_SM(ratesArray[rate11s], 24)
              | POW_SM(ratesArray[rate11l], 16)
              | POW_SM(ratesArray[rate5_5s],  8)
              | POW_SM(ratesArray[rate5_5l],  0)
        );
    }

    /* Write the HT20 power per rate set */
    OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE5,
        POW_SM(ratesArray[rateHt20_3], 24)
          | POW_SM(ratesArray[rateHt20_2],  16)
          | POW_SM(ratesArray[rateHt20_1],  8)
          | POW_SM(ratesArray[rateHt20_0],   0)
    );
    OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE6,
        POW_SM(ratesArray[rateHt20_7], 24)
          | POW_SM(ratesArray[rateHt20_6],  16)
          | POW_SM(ratesArray[rateHt20_5],  8)
          | POW_SM(ratesArray[rateHt20_4],   0)
    );

    if (ht && ht->ht_phymode == HAL_HT_PHYMODE_2040) {
        /* Write the HT40 power per rate set */
        OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE7,
            POW_SM(ratesArray[rateHt40_3], 24)
              | POW_SM(ratesArray[rateHt40_2],  16)
              | POW_SM(ratesArray[rateHt40_1],  8)
              | POW_SM(ratesArray[rateHt40_0],   0)
        );
        OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE8,
            POW_SM(ratesArray[rateHt40_7], 24)
              | POW_SM(ratesArray[rateHt40_6],  16)
              | POW_SM(ratesArray[rateHt40_5],  8)
              | POW_SM(ratesArray[rateHt40_4],   0)
        );
        /* Write the Dup/Ext 40 power per rate set */
        OS_REG_WRITE(ah, AR_PHY_POWER_TX_RATE9,
            POW_SM(ratesArray[rateExtOfdm], 24)
              | POW_SM(ratesArray[rateExtCck],  16)
              | POW_SM(ratesArray[rateDupOfdm],  8)
              | POW_SM(ratesArray[rateDupCck],   0)
        );
    }

    /* Write the Power subtraction for dynamic chain changing */
    OS_REG_WRITE(ah, AR_PHY_POWER_TX_SUB,
        POW_SM(pModal->pwrDecreaseFor3Chain, 6)
          | POW_SM(pModal->pwrDecreaseFor2Chain, 0)
    );
    return AH_TRUE;
}

/***************************************
 * Helper functions common for AP/CB/XB
 **************************************/

/**************************************************************
 * ar5416SetPowerPerRateTable
 *
 * Sets the transmit power in the baseband for the given
 * operating channel and mode.
 */
static  HAL_BOOL
ar5416SetPowerPerRateTable(struct ath_hal *ah, ar5416_eeprom_t *pEepData,
                           HAL_HT *ht, HAL_CHANNEL_INTERNAL *chan,
                           int16_t *ratesArray, u_int16_t cfgCtl,
                           u_int16_t AntennaReduction, 
                           u_int16_t twiceMaxRegulatoryPower,
                           u_int16_t powerLimit)
{
/* Local defines to distinguish between extension and control CTL's */
#define EXT_ADDITIVE (0x8000)
#define CTL_11A_EXT (CTL_11A | EXT_ADDITIVE)
#define CTL_11G_EXT (CTL_11G | EXT_ADDITIVE)
#define CTL_11B_EXT (CTL_11B | EXT_ADDITIVE)
    u_int16_t twiceMaxEdgePower = AR5416_MAX_RATE_POWER;
    int i;
    int16_t  twiceLargestAntenna;
    CAL_CTL_DATA *rep;
    CAL_TARGET_POWER_LEG targetPowerOfdm, targetPowerCck = {0, {0, 0, 0, 0}};
    CAL_TARGET_POWER_LEG targetPowerOfdmExt = {0, {0, 0, 0, 0}}, targetPowerCckExt = {0, {0, 0, 0, 0}};
    CAL_TARGET_POWER_HT  targetPowerHt20, targetPowerHt40 = {0, {0, 0, 0, 0}};
    int16_t scaledPower, minCtlPower;
    u_int16_t ctlModesFor11a[] = {CTL_11A, CTL_5GHT20, CTL_5GHT40};
    u_int16_t ctlModesFor11g[] = {CTL_11B, CTL_11G, CTL_2GHT20, CTL_2GHT40};
    u_int16_t numCtlModes, *pCtlMode, ctlMode, freq;
    CHAN_CENTERS centers;

    ar5416GetChannelCenters(ah, ht, chan, &centers);

    /* Compute TxPower reduction due to Antenna Gain */
    twiceLargestAntenna = AH_MAX(AH_MAX(pEepData->modalHeader[IS_CHAN_2GHZ(chan)].antennaGainCh[0], pEepData->modalHeader[IS_CHAN_2GHZ(chan)].antennaGainCh[1]),
        pEepData->modalHeader[IS_CHAN_2GHZ(chan)].antennaGainCh[0]);
    twiceLargestAntenna =  (int16_t)AH_MIN((AntennaReduction) - twiceLargestAntenna, 0);

    /* scaledPower is the minimum of the user input power level and the regulatory allowed power level */
    scaledPower = AH_MIN(powerLimit, twiceMaxRegulatoryPower + twiceLargestAntenna);

    /* Reduce scaled Power by number of chains active to get to per chain tx power level */
    /* TODO: better value than these? */
    switch (owl_get_ntxchains(ath_hal_tx_chainmask)) {
    case 1:
        break;
    case 2:
        scaledPower -= pEepData->modalHeader[IS_CHAN_2GHZ(chan)].pwrDecreaseFor2Chain;
        break;
    case 3:
        scaledPower -= pEepData->modalHeader[IS_CHAN_2GHZ(chan)].pwrDecreaseFor3Chain;
        break;
    default:
        HALASSERT(0); /* Unsupported number of chains */
    }

    scaledPower = AH_MAX(0, scaledPower);

    /* Get target powers from EEPROM - our baseline for TX Power */
    if (IS_CHAN_2GHZ(chan)) {
        /* Setup for CTL modes */
        numCtlModes = N(ctlModesFor11g) - 3; // No ExtCck, ExtOfd, HT40
        pCtlMode = ctlModesFor11g;

        ar5416GetTargetPowersLeg(ah, ht, chan, pEepData->calTargetPowerCck,
                              AR5416_NUM_2G_CCK_TARGET_POWERS, &targetPowerCck, 4, AH_FALSE);
        ar5416GetTargetPowersLeg(ah, ht, chan, pEepData->calTargetPower2G,
                              AR5416_NUM_2G_20_TARGET_POWERS, &targetPowerOfdm, 4, AH_FALSE);
        ar5416GetTargetPowers(ah, ht, chan, pEepData->calTargetPower2GHT20,
                              AR5416_NUM_2G_20_TARGET_POWERS, &targetPowerHt20, 8, AH_FALSE);
        if (ht && ht->ht_phymode == HAL_HT_PHYMODE_2040) {
            numCtlModes = N(ctlModesFor11g);
            ar5416GetTargetPowers(ah, ht, chan, pEepData->calTargetPower2GHT40,
                                  AR5416_NUM_2G_40_TARGET_POWERS, &targetPowerHt40, 8, AH_TRUE);
            ar5416GetTargetPowersLeg(ah, ht, chan, 
                                     pEepData->calTargetPowerCck,
                                     AR5416_NUM_2G_CCK_TARGET_POWERS, 
                                     &targetPowerCckExt, 4, AH_TRUE);
            ar5416GetTargetPowersLeg(ah, ht, chan, pEepData->calTargetPower2G,
                                  AR5416_NUM_2G_20_TARGET_POWERS, &targetPowerOfdmExt, 4, AH_TRUE);
        }
    } else {
        /* Setup for CTL modes */
        numCtlModes = N(ctlModesFor11a)  - 2; // No Ext or HT40
        pCtlMode = ctlModesFor11a;

        ar5416GetTargetPowersLeg(ah, ht, chan, pEepData->calTargetPower5G,
                              AR5416_NUM_5G_20_TARGET_POWERS, &targetPowerOfdm, 4, AH_FALSE);
        ar5416GetTargetPowers(ah, ht, chan, pEepData->calTargetPower5GHT20,
                              AR5416_NUM_2G_20_TARGET_POWERS, &targetPowerHt20, 8, AH_FALSE);
        if (ht && ht->ht_phymode == HAL_HT_PHYMODE_2040) {
            numCtlModes = N(ctlModesFor11a); // No Ext or HT40
            ar5416GetTargetPowers(ah, ht, chan, pEepData->calTargetPower5GHT40,
                                  AR5416_NUM_2G_40_TARGET_POWERS, &targetPowerHt40, 8, AH_TRUE);
            ar5416GetTargetPowersLeg(ah, ht, chan, pEepData->calTargetPower5G,
                                  AR5416_NUM_5G_20_TARGET_POWERS, &targetPowerOfdmExt, 4, AH_TRUE);
        }

    }

    /* For MIMO, need to apply regulatory caps individually across dynamically running modes: CCK, OFDM, HT20, HT40 */
    for (ctlMode = 0; ctlMode < numCtlModes; ctlMode++) {
        HAL_BOOL isHt40CtlMode = (pCtlMode[ctlMode] == CTL_5GHT40) || (pCtlMode[ctlMode] == CTL_2GHT40);
        if (isHt40CtlMode) {
            freq = centers.ctl_center;
        } else if (pCtlMode[ctlMode] & EXT_ADDITIVE) {
            freq = centers.ext_center;
        } else {
            freq = centers.ctl_center;
        }
        for (i = 0; (i < AR5416_NUM_CTLS) && pEepData->ctlIndex[i]; i++) {
            u_int16_t twiceMinEdgePower;

            if (((cfgCtl | (pCtlMode[ctlMode] & CTL_MODE_M)) == pEepData->ctlIndex[i]) ||
                ((cfgCtl | (pCtlMode[ctlMode] & CTL_MODE_M))== ((pEepData->ctlIndex[i] & CTL_MODE_M) | SD_NO_CTL)))
            {
                rep = &(pEepData->ctlData[i]);
                twiceMinEdgePower = ar5416GetMaxEdgePower(freq, rep->ctlEdges[owl_get_ntxchains(ath_hal_tx_chainmask)], IS_CHAN_2GHZ(chan));
                if (cfgCtl == SD_NO_CTL) {
                    /* Find the minimum of all CTL edge powers that apply to this channel */
                    twiceMaxEdgePower = AH_MIN(twiceMaxEdgePower, twiceMinEdgePower);
                } else {
                    twiceMaxEdgePower = twiceMinEdgePower;
                    break;
                }
            }
        }
        minCtlPower = (u_int8_t)AH_MIN(twiceMaxEdgePower, scaledPower);
        /* ==Apply ctl mode to correct target power set== */
        switch(pCtlMode[ctlMode]) {
        case CTL_11B:
            for (i = 0; i < N(targetPowerCck.tPow2x); i++) {
                targetPowerCck.tPow2x[i] = (u_int8_t)AH_MIN(targetPowerCck.tPow2x[i], minCtlPower);
            }
            break;
        case CTL_11A:
        case CTL_11G:
            for (i = 0; i < N(targetPowerOfdm.tPow2x); i++) {
                targetPowerOfdm.tPow2x[i] = (u_int8_t)AH_MIN(targetPowerOfdm.tPow2x[i], minCtlPower);
            }
            break;
        case CTL_5GHT20:
        case CTL_2GHT20:
            for (i = 0; i < N(targetPowerHt20.tPow2x); i++) {
                targetPowerHt20.tPow2x[i] = (u_int8_t)AH_MIN(targetPowerHt20.tPow2x[i], minCtlPower);
            }
            break;
        case CTL_11B_EXT:
            targetPowerCckExt.tPow2x[0] = (u_int8_t)AH_MIN(targetPowerCckExt.tPow2x[0], minCtlPower);
            break;
        case CTL_11A_EXT:
        case CTL_11G_EXT:
            targetPowerOfdmExt.tPow2x[0] = (u_int8_t)AH_MIN(targetPowerOfdmExt.tPow2x[0], minCtlPower);
            break;
        case CTL_5GHT40:
        case CTL_2GHT40:
            for (i = 0; i < N(targetPowerHt40.tPow2x); i++) {
                targetPowerHt40.tPow2x[i] = (u_int8_t)AH_MIN(targetPowerHt40.tPow2x[i], minCtlPower);
            }
            break;
        default:
            HALASSERT(0);
            break;
        }

    } /* end ctl mode checking */

    /* Set rates Array from collected data */
    ratesArray[rate6mb] = ratesArray[rate9mb] = ratesArray[rate12mb] = ratesArray[rate18mb] = ratesArray[rate24mb] = targetPowerOfdm.tPow2x[0];
    ratesArray[rate36mb] = targetPowerOfdm.tPow2x[1];
    ratesArray[rate48mb] = targetPowerOfdm.tPow2x[2];
    ratesArray[rate54mb] = targetPowerOfdm.tPow2x[3];
    ratesArray[rateXr] = targetPowerOfdm.tPow2x[0];

    for (i = 0; i < N(targetPowerHt20.tPow2x); i++) {
        ratesArray[rateHt20_0 + i] = targetPowerHt20.tPow2x[i];
    }

    if (IS_CHAN_2GHZ(chan)) {
        ratesArray[rate1l]  = targetPowerCck.tPow2x[0];
        ratesArray[rate2s] = ratesArray[rate2l]  = targetPowerCck.tPow2x[1];
        ratesArray[rate5_5s] = ratesArray[rate5_5l] = targetPowerCck.tPow2x[2];;
        ratesArray[rate11s] = ratesArray[rate11l] = targetPowerCck.tPow2x[3];;
    }
    if (ht && ht->ht_phymode == HAL_HT_PHYMODE_2040) {
        for (i = 0; i < N(targetPowerHt40.tPow2x); i++) {
            ratesArray[rateHt40_0 + i] = targetPowerHt40.tPow2x[i];
        }
        ratesArray[rateDupOfdm] = targetPowerHt40.tPow2x[0];
        ratesArray[rateDupCck]  = targetPowerHt40.tPow2x[0];
        ratesArray[rateExtOfdm] = targetPowerOfdmExt.tPow2x[0];
        if (IS_CHAN_2GHZ(chan)) {
            ratesArray[rateExtCck]  = targetPowerCckExt.tPow2x[0];
        }
    }
    return AH_TRUE;
#undef EXT_ADDITIVE
#undef CTL_11A_EXT
#undef CTL_11G_EXT
#undef CTL_11B_EXT
}

/**************************************************************
 * ar5416GetMaxEdgePower
 *
 * Find the maximum conformance test limit for the given channel and CTL info
 */
static u_int16_t
ar5416GetMaxEdgePower(u_int16_t freq, CAL_CTL_EDGES *pRdEdgesPower, HAL_BOOL is2GHz)
{
    u_int16_t twiceMaxEdgePower = AR5416_MAX_RATE_POWER;
    int      i;

    /* Get the edge power */
    for (i = 0; (i < AR5416_NUM_BAND_EDGES) && (pRdEdgesPower[i].bChannel != AR5416_BCHAN_UNUSED) ; i++) {
        /*
         * If there's an exact channel match or an inband flag set
         * on the lower channel use the given rdEdgePower
         */
        if (freq == fbin2freq(pRdEdgesPower[i].bChannel, is2GHz)) {
            twiceMaxEdgePower = pRdEdgesPower[i].tPower;
            break;
        } else if ((i > 0) && (freq < fbin2freq(pRdEdgesPower[i].bChannel, is2GHz))) {
            if (fbin2freq(pRdEdgesPower[i - 1].bChannel, is2GHz) < freq && pRdEdgesPower[i - 1].flag) {
                twiceMaxEdgePower = pRdEdgesPower[i - 1].tPower;
            }
            /* Leave loop - no more affecting edges possible in this monotonic increasing list */
            break;
        }
    }
    HALASSERT(twiceMaxEdgePower > 0);
    return twiceMaxEdgePower;
}

/**************************************************************
 * ar5416GetTargetPowers
 *
 * Return the four rates of target power for the given target power table
 * channel, and number of channels
 */
static void
ar5416GetTargetPowers(struct ath_hal *ah, HAL_HT *ht, HAL_CHANNEL_INTERNAL *chan,
                      CAL_TARGET_POWER_HT *powInfo, u_int16_t numChannels,
                      CAL_TARGET_POWER_HT *pNewPower, u_int16_t numRates,
                      HAL_BOOL isHt40Target)
{
    u_int16_t clo, chi;
    int i;
    int matchIndex = -1, lowIndex = -1;
    u_int16_t freq;
    CHAN_CENTERS centers;

    ar5416GetChannelCenters(ah, ht, chan, &centers);
    freq = isHt40Target ? centers.synth_center : centers.ctl_center;

    /* Copy the target powers into the temp channel list */
    if (freq <= fbin2freq(powInfo[0].bChannel, IS_CHAN_2GHZ(chan))) {
        matchIndex = 0;
    } else {
        for (i = 0; (i < numChannels) && (powInfo[i].bChannel != AR5416_BCHAN_UNUSED); i++) {
            if (freq == fbin2freq(powInfo[i].bChannel, IS_CHAN_2GHZ(chan))) {
                matchIndex = i;
                break;
            } else if ((freq < fbin2freq(powInfo[i].bChannel, IS_CHAN_2GHZ(chan))) &&
                       (freq > fbin2freq(powInfo[i - 1].bChannel, IS_CHAN_2GHZ(chan))))
            {
                lowIndex = i - 1;
                break;
            }
        }
        if ((matchIndex == -1) && (lowIndex == -1)) {
            HALASSERT(freq > fbin2freq(powInfo[i - 1].bChannel, IS_CHAN_2GHZ(chan)));
            matchIndex = i - 1;
        }
    }

    if (matchIndex != -1) {
        *pNewPower = powInfo[matchIndex];
    } else {
        HALASSERT(lowIndex != -1);
        /*
         * Get the lower and upper channels, target powers,
         * and interpolate between them.
         */
        clo = fbin2freq(powInfo[lowIndex].bChannel, IS_CHAN_2GHZ(chan));
        chi = fbin2freq(powInfo[lowIndex + 1].bChannel, IS_CHAN_2GHZ(chan));

        for (i = 0; i < numRates; i++) {
            pNewPower->tPow2x[i] = (u_int8_t)interpolate(freq, clo, chi,
                                   powInfo[lowIndex].tPow2x[i], powInfo[lowIndex + 1].tPow2x[i]);
        }
    }
}

static void
ar5416GetTargetPowersLeg(struct ath_hal *ah, HAL_HT *ht,
                         HAL_CHANNEL_INTERNAL *chan,
                         CAL_TARGET_POWER_LEG *powInfo, u_int16_t numChannels,
                         CAL_TARGET_POWER_LEG *pNewPower, u_int16_t numRates,
			 HAL_BOOL isExtTarget)
{
    u_int16_t clo, chi;
    int i;
    int matchIndex = -1, lowIndex = -1;
    u_int16_t freq;
    CHAN_CENTERS centers;

    ar5416GetChannelCenters(ah, ht, chan, &centers);
    freq = (isExtTarget) ? centers.ext_center :centers.ctl_center;

    /* Copy the target powers into the temp channel list */
    if (freq <= fbin2freq(powInfo[0].bChannel, IS_CHAN_2GHZ(chan))) {
        matchIndex = 0;
    } else {
        for (i = 0; (i < numChannels) && (powInfo[i].bChannel != AR5416_BCHAN_UNUSED); i++) {
            if (freq == fbin2freq(powInfo[i].bChannel, IS_CHAN_2GHZ(chan))) {
                matchIndex = i;
                break;
            } else if ((freq < fbin2freq(powInfo[i].bChannel, IS_CHAN_2GHZ(chan))) &&
                       (freq > fbin2freq(powInfo[i - 1].bChannel, IS_CHAN_2GHZ(chan))))
            {
                lowIndex = i - 1;
                break;
            }
        }
        if ((matchIndex == -1) && (lowIndex == -1)) {
            HALASSERT(freq > fbin2freq(powInfo[i - 1].bChannel, IS_CHAN_2GHZ(chan)));
            matchIndex = i - 1;
        }
    }

    if (matchIndex != -1) {
        *pNewPower = powInfo[matchIndex];
    } else {
        HALASSERT(lowIndex != -1);
        /*
         * Get the lower and upper channels, target powers,
         * and interpolate between them.
         */
        clo = fbin2freq(powInfo[lowIndex].bChannel, IS_CHAN_2GHZ(chan));
        chi = fbin2freq(powInfo[lowIndex + 1].bChannel, IS_CHAN_2GHZ(chan));

        for (i = 0; i < numRates; i++) {
            pNewPower->tPow2x[i] = (u_int8_t)interpolate(freq, clo, chi,
                                   powInfo[lowIndex].tPow2x[i], powInfo[lowIndex + 1].tPow2x[i]);
        }
    }
}

/**************************************************************
 * ar5416SetPowerCalTable
 *
 * Pull the PDADC piers from cal data and interpolate them across the given
 * points as well as from the nearest pier(s) to get a power detector
 * linear voltage to power level table.
 */
static HAL_BOOL
ar5416SetPowerCalTable(struct ath_hal *ah, HAL_HT *ht, ar5416_eeprom_t *pEepData, HAL_CHANNEL_INTERNAL *chan, int16_t *pTxPowerIndexOffset)
{
    CAL_DATA_PER_FREQ *pRawDataset;
    u_int8_t  *pCalBChans = AH_NULL;
    u_int16_t pdGainOverlap_t2;
    static u_int8_t  pdadcValues[AR5416_NUM_PDADC_VALUES];
    u_int16_t gainBoundaries[AR5416_PD_GAINS_IN_MASK];
    u_int16_t numPiers, i, j;
    int16_t  tMinCalPower;
    u_int16_t numXpdGain, xpdMask;
    u_int16_t xpdGainValues[AR5416_NUM_PD_GAINS];
    u_int32_t reg32, regOffset, regChainOffset;

    pdGainOverlap_t2 = (u_int16_t)(MS(OS_REG_READ(ah, AR_PHY_TPCRG5), AR_PHY_TPCRG5_PD_GAIN_OVERLAP));

    if (IS_CHAN_2GHZ(chan)) {
        pCalBChans = pEepData->calFreqPier2G;
        numPiers = AR5416_NUM_2G_CAL_PIERS;
        xpdMask = pEepData->modalHeader[1].xpdGain;
    } else {
        pCalBChans = pEepData->calFreqPier5G;
        numPiers = AR5416_NUM_5G_CAL_PIERS;
        xpdMask = pEepData->modalHeader[0].xpdGain;
    }

    numXpdGain = 0;
    /* Calculate the value of xpdgains from the xpdGain Mask */
    for (i = 1; i <= AR5416_PD_GAINS_IN_MASK; i++) {
        if ((xpdMask >> (AR5416_PD_GAINS_IN_MASK - i)) & 1) {
            if (numXpdGain >= AR5416_NUM_PD_GAINS) {
                HALASSERT(0);
                break;
            }
            xpdGainValues[numXpdGain] = (u_int16_t)(AR5416_PD_GAINS_IN_MASK - i);
            numXpdGain++;
        }
    }

    OS_REG_RMW_FIELD(ah, AR_PHY_TPCRG1, AR_PHY_TPCRG1_NUM_PD_GAIN, (numXpdGain - 1) & 0x3);

    for (i = 0; i < AR5416_MAX_CHAINS; i++) {

            if ( !(IS_5416V1(ah) ) &&
	    //if ((AH_PRIVATE(ah)->ah_macRev >= AR_SREV_REVISION_OWL_20) && 
            (pEepData->baseEepHeader.txMask == 0x5) && (i != 0)) {
            /* Regs are swapped from chain 2 to 1 for 5416 2_0 with 
             * only chains 0 and 2 populated 
             */
            regChainOffset = (i == 1) ? 0x2000 : 0x1000;
        } else {
            regChainOffset = i * 0x1000;
        }

        if (pEepData->baseEepHeader.txMask & (1 << i)) {
            if (IS_CHAN_2GHZ(chan)) {
                pRawDataset = pEepData->calPierData2G[i];
            } else {
                pRawDataset = pEepData->calPierData5G[i];
            }

            ar5416GetGainBoundariesAndPdadcs(ah, ht, chan, pRawDataset,
                                             pCalBChans, numPiers,
                                             pdGainOverlap_t2,
                                             &tMinCalPower, gainBoundaries,
                                             pdadcValues, numXpdGain);

            if ((i == 0) || !(IS_5416V1(ah))) {
	    	//(AH_PRIVATE(ah)->ah_macRev >= AR_SREV_REVISION_OWL_20)) {
                /*
                 * Note the pdadc table may not start at 0 dBm power, could be
                 * negative or greater than 0.  Need to offset the power
                 * values by the amount of minPower for griffin
                 */

                OS_REG_WRITE(ah, AR_PHY_TPCRG5 + regChainOffset,
                     SM(pdGainOverlap_t2, AR_PHY_TPCRG5_PD_GAIN_OVERLAP) |
                     SM(gainBoundaries[0], AR_PHY_TPCRG5_PD_GAIN_BOUNDARY_1)  |
                     SM(gainBoundaries[1], AR_PHY_TPCRG5_PD_GAIN_BOUNDARY_2)  |
                     SM(gainBoundaries[2], AR_PHY_TPCRG5_PD_GAIN_BOUNDARY_3)  |
                     SM(gainBoundaries[3], AR_PHY_TPCRG5_PD_GAIN_BOUNDARY_4));

//printf("SNOOP: a26c = %x\n",
//                     (pdGainOverlap_t2 & 0xf) |
//                     ((gainBoundaries[0] & 0x3f) << 4)  |
//                     ((gainBoundaries[1] & 0x3f) << 10) |
//                     ((gainBoundaries[2] & 0x3f) << 16) |
//                     ((gainBoundaries[3] & 0x3f) << 22));

            }

            /* Write the power values into the baseband power table */
            regOffset = AR_PHY_BASE + (672 << 2) + regChainOffset;
            for (j = 0; j < 32; j++) {
                reg32 = ((pdadcValues[4*j + 0] & 0xFF) << 0)  |
                    ((pdadcValues[4*j + 1] & 0xFF) << 8)  |
                    ((pdadcValues[4*j + 2] & 0xFF) << 16) |
                    ((pdadcValues[4*j + 3] & 0xFF) << 24) ;
                OS_REG_WRITE(ah, regOffset, reg32);
//printf("Reg %x = %x\n", regOffset, reg32);
                regOffset += 4;
            }
        }
    }
    *pTxPowerIndexOffset = 0;

    return AH_TRUE;
}

/**************************************************************
 * ar5416GetGainBoundariesAndPdadcs
 *
 * Uses the data points read from EEPROM to reconstruct the pdadc power table
 * Called by ar5416SetPowerCalTable only.
 */
static void
ar5416GetGainBoundariesAndPdadcs(struct ath_hal *ah, HAL_HT *ht,
                                 HAL_CHANNEL_INTERNAL *chan, CAL_DATA_PER_FREQ *pRawDataSet,
                                 u_int8_t * bChans,  u_int16_t availPiers,
                                 u_int16_t tPdGainOverlap, int16_t *pMinCalPower, u_int16_t * pPdGainBoundaries,
                                 u_int8_t * pPDADCValues, u_int16_t numXpdGains)
{

    int       i, j, k;
    int16_t   ss;         /* potentially -ve index for taking care of pdGainOverlap */
    u_int16_t  idxL, idxR, numPiers; /* Pier indexes */

    /* filled out Vpd table for all pdGains (chanL) */
    static u_int8_t   vpdTableL[AR5416_NUM_PD_GAINS][AR5416_MAX_PWR_RANGE_IN_HALF_DB];

    /* filled out Vpd table for all pdGains (chanR) */
    static u_int8_t   vpdTableR[AR5416_NUM_PD_GAINS][AR5416_MAX_PWR_RANGE_IN_HALF_DB];

    /* filled out Vpd table for all pdGains (interpolated) */
    static u_int8_t   vpdTableI[AR5416_NUM_PD_GAINS][AR5416_MAX_PWR_RANGE_IN_HALF_DB];

    u_int8_t   *pVpdL, *pVpdR, *pPwrL, *pPwrR;
    u_int8_t   minPwrT4[AR5416_NUM_PD_GAINS];
    u_int8_t   maxPwrT4[AR5416_NUM_PD_GAINS];
    int16_t   vpdStep;
    int16_t   tmpVal;
    u_int16_t  sizeCurrVpdTable, maxIndex, tgtIndex;
    HAL_BOOL    match;
    int16_t  minDelta = 0;
    CHAN_CENTERS centers;

    ar5416GetChannelCenters(ah, ht, chan, &centers);

    /* Trim numPiers for the number of populated channel Piers */
    for (numPiers = 0; numPiers < availPiers; numPiers++) {
        if (bChans[numPiers] == AR5416_BCHAN_UNUSED) {
            break;
        }
    }

    /* Find pier indexes around the current channel */
    match = getLowerUpperIndex((u_int8_t)FREQ2FBIN(centers.synth_center, IS_CHAN_2GHZ(chan)),
        bChans, numPiers, &idxL,
        &idxR);

    if (match) {
        /* Directly fill both vpd tables from the matching index */
        for (i = 0; i < numXpdGains; i++) {
            minPwrT4[i] = pRawDataSet[idxL].pwrPdg[i][0];
            maxPwrT4[i] = pRawDataSet[idxL].pwrPdg[i][4];
            ar5416FillVpdTable(minPwrT4[i], maxPwrT4[i], pRawDataSet[idxL].pwrPdg[i],
                               pRawDataSet[idxL].vpdPdg[i], AR5416_PD_GAIN_ICEPTS, vpdTableI[i]);
        }
    } else {
        for (i = 0; i < numXpdGains; i++) {
            pVpdL = pRawDataSet[idxL].vpdPdg[i];
            pPwrL = pRawDataSet[idxL].pwrPdg[i];
            pVpdR = pRawDataSet[idxR].vpdPdg[i];
            pPwrR = pRawDataSet[idxR].pwrPdg[i];

            /* Start Vpd interpolation from the max of the minimum powers */
            minPwrT4[i] = AH_MAX(pPwrL[0], pPwrR[0]);

            /* End Vpd interpolation from the min of the max powers */
            maxPwrT4[i] = AH_MIN(pPwrL[AR5416_PD_GAIN_ICEPTS - 1], pPwrR[AR5416_PD_GAIN_ICEPTS - 1]);
            HALASSERT(maxPwrT4[i] > minPwrT4[i]);

            /* Fill pier Vpds */
            ar5416FillVpdTable(minPwrT4[i], maxPwrT4[i], pPwrL, pVpdL, AR5416_PD_GAIN_ICEPTS, vpdTableL[i]);
            ar5416FillVpdTable(minPwrT4[i], maxPwrT4[i], pPwrR, pVpdR, AR5416_PD_GAIN_ICEPTS, vpdTableR[i]);

            /* Interpolate the final vpd */
            for (j = 0; j < (maxPwrT4[i] - minPwrT4[i]) / 2; j++) {
                vpdTableI[i][j] = (u_int8_t)(interpolate((u_int16_t)FREQ2FBIN(centers.synth_center, IS_CHAN_2GHZ(chan)),
                    bChans[idxL], bChans[idxR], vpdTableL[i][j], vpdTableR[i][j]));
            }
        }
    }
    *pMinCalPower = (int16_t)(minPwrT4[0] / 2);

    k = 0; /* index for the final table */
    for (i = 0; i < numXpdGains; i++) {
        if (i == (numXpdGains - 1)) {
            pPdGainBoundaries[i] = (u_int16_t)(maxPwrT4[i] / 2);
        } else {
            pPdGainBoundaries[i] = (u_int16_t)((maxPwrT4[i] + minPwrT4[i+1]) / 4);
        }

        pPdGainBoundaries[i] = (u_int16_t)AH_MIN(AR5416_MAX_RATE_POWER, pPdGainBoundaries[i]);

        //WORKAROUND for OWL 1.0 until we get a per chain gain boundary register.
        //this is not the best solution
        if ((i == 0) && IS_5416V1(ah) ) {
	    //!(AH_PRIVATE(ah)->ah_macRev >= AR_SREV_REVISION_OWL_20)){
            //fix the gain delta, but get a delta that can be applied to min to
            //keep the upper power values accurate, don't think max needs to
            //be adjusted because should not be at that area of the table?
            minDelta = pPdGainBoundaries[0] - 23;
            pPdGainBoundaries[0] = 23;
        }
        else {
            minDelta = 0;
        }

        /* Find starting index for this pdGain */
        if (i == 0) {
            ss = 0; /* for the first pdGain, start from index 0 */
        } else {
            ss = (int16_t)((pPdGainBoundaries[i-1] - (minPwrT4[i] / 2)) - tPdGainOverlap + 1 + minDelta);
        }
        vpdStep = (int16_t)(vpdTableI[i][1] - vpdTableI[i][0]);
        vpdStep = (int16_t)((vpdStep < 1) ? 1 : vpdStep);
        /*
         *-ve ss indicates need to extrapolate data below for this pdGain
         */
        while ((ss < 0) && (k < (AR5416_NUM_PDADC_VALUES - 1))) {
            tmpVal = (int16_t)(vpdTableI[i][0] + ss * vpdStep);
            pPDADCValues[k++] = (u_int8_t)((tmpVal < 0) ? 0 : tmpVal);
            ss++;
        }

        sizeCurrVpdTable = (u_int8_t)((maxPwrT4[i] - minPwrT4[i]) / 2);
        tgtIndex = (u_int8_t)(pPdGainBoundaries[i] + tPdGainOverlap - (minPwrT4[i] / 2));
        maxIndex = (tgtIndex < sizeCurrVpdTable) ? tgtIndex : sizeCurrVpdTable;

        while ((ss < maxIndex) && (k < (AR5416_NUM_PDADC_VALUES - 1))) {
            pPDADCValues[k++] = vpdTableI[i][ss++];
        }

        vpdStep = (int16_t)(vpdTableI[i][sizeCurrVpdTable - 1] - vpdTableI[i][sizeCurrVpdTable - 2]);
        vpdStep = (int16_t)((vpdStep < 1) ? 1 : vpdStep);
        /*
         * for last gain, pdGainBoundary == Pmax_t2, so will
         * have to extrapolate
         */
        if (tgtIndex > maxIndex) {  /* need to extrapolate above */
            while ((ss <= tgtIndex) && (k < (AR5416_NUM_PDADC_VALUES - 1))) {
                tmpVal = (int16_t)((vpdTableI[i][sizeCurrVpdTable - 1] +
                          (ss - maxIndex) * vpdStep));
                pPDADCValues[k++] = (u_int8_t)((tmpVal > 255) ? 255 : tmpVal);
                ss++;
            }
        }               /* extrapolated above */
    }                   /* for all pdGainUsed */

    /* Fill out pdGainBoundaries - only up to 2 allowed here, but hardware allows up to 4 */
    while (i < AR5416_PD_GAINS_IN_MASK) {
        pPdGainBoundaries[i] = pPdGainBoundaries[i-1];
        i++;
    }

    while (k < AR5416_NUM_PDADC_VALUES) {
        pPDADCValues[k] = pPDADCValues[k-1];
        k++;
    }
    return;
}

/**************************************************************
 * getLowerUppderIndex
 *
 * Return indices surrounding the value in sorted integer lists.
 * Requirement: the input list must be monotonically increasing
 *     and populated up to the list size
 * Returns: match is set if an index in the array matches exactly
 *     or a the target is before or after the range of the array.
 */
 HAL_BOOL
getLowerUpperIndex(u_int8_t target, u_int8_t *pList, u_int16_t listSize,
                   u_int16_t *indexL, u_int16_t *indexR)
{
    u_int16_t i;

    /*
     * Check first and last elements for beyond ordered array cases.
     */
    if (target <= pList[0]) {
        *indexL = *indexR = 0;
        return AH_TRUE;
    }
    if (target >= pList[listSize-1]) {
        *indexL = *indexR = (u_int16_t)(listSize - 1);
        return AH_TRUE;
    }

    /* look for value being near or between 2 values in list */
    for (i = 0; i < listSize - 1; i++) {
        /*
         * If value is close to the current value of the list
         * then target is not between values, it is one of the values
         */
        if (pList[i] == target) {
            *indexL = *indexR = i;
            return AH_TRUE;
        }
        /*
         * Look for value being between current value and next value
         * if so return these 2 values
         */
        if (target < pList[i + 1]) {
            *indexL = i;
            *indexR = (u_int16_t)(i + 1);
            return AH_FALSE;
        }
    }
    HALASSERT(0);
    return AH_FALSE;
}

/**************************************************************
 * ar5416FillVpdTable
 *
 * Fill the Vpdlist for indices Pmax-Pmin
 * Note: pwrMin, pwrMax and Vpdlist are all in dBm * 4
 */
static HAL_BOOL
ar5416FillVpdTable(u_int8_t pwrMin, u_int8_t pwrMax, u_int8_t *pPwrList,
                   u_int8_t *pVpdList, u_int16_t numIntercepts, u_int8_t *pRetVpdList)
{
    u_int16_t  i, k;
    u_int8_t   currPwr = pwrMin;
    u_int16_t  idxL, idxR;

    HALASSERT(pwrMax > pwrMin);
    for (i = 0; i <= (pwrMax - pwrMin) / 2; i++) {
        getLowerUpperIndex(currPwr, pPwrList, numIntercepts,
                           &(idxL), &(idxR));
        if (idxR < 1)
            idxR = 1;           /* extrapolate below */
        if (idxL == numIntercepts - 1)
            idxL = (u_int16_t)(numIntercepts - 2);   /* extrapolate above */
        if (pPwrList[idxL] == pPwrList[idxR])
            k = pVpdList[idxL];
        else
            k = (u_int16_t)( ((currPwr - pPwrList[idxL]) * pVpdList[idxR] + (pPwrList[idxR] - currPwr) * pVpdList[idxL]) /
                  (pPwrList[idxR] - pPwrList[idxL]) );
        HALASSERT(k < 256);
        pRetVpdList[i] = (u_int8_t)k;
        currPwr += 2;               /* half dB steps */
    }

    return AH_TRUE;
}

/**************************************************************************
 * interpolate
 *
 * Returns signed interpolated or the scaled up interpolated value
 */
static int16_t
interpolate(u_int16_t target, u_int16_t srcLeft, u_int16_t srcRight,
            int16_t targetLeft, int16_t targetRight)
{
    int16_t rv;

    if (srcRight == srcLeft) {
        rv = targetLeft;
    } else {
        rv = (int16_t)( ((target - srcLeft) * targetRight +
              (srcRight - target) * targetLeft) / (srcRight - srcLeft) );
    }
    return rv;
}

/**************************************************************************
 * fbin2freq
 *
 * Get channel value from binary representation held in eeprom
 * RETURNS: the frequency in MHz
 */
static u_int16_t
fbin2freq(u_int8_t fbin, HAL_BOOL is2GHz)
{
    /*
     * Reserved value 0xFF provides an empty definition both as
     * an fbin and as a frequency - do not convert
     */
    if (fbin == AR5416_BCHAN_UNUSED) {
        return fbin;
    }

    return (u_int16_t)((is2GHz) ? (2300 + fbin) : (4800 + 5 * fbin));
}

static HAL_STATUS
ar5416CheckEeprom(struct ath_hal *ah)
{
    u_int32_t sum,len = 0;
    u_int16_t *eepdata;
    int i;
    struct ath_hal_5416 *ahp = AH5416(ah);
    ar5416_eeprom_t *pEeprom = (ar5416_eeprom_t *)ahp->ah_5416eeprom;

    /* eeprom size is hardcoded to 4k */
#if 0
    if (!owl_eeprom_read(ah, AR_EEPROM_PROTECT, &eeval)) {
        HALDEBUG(ah, "%s: cannot read EEPROM protection "
            "bits; read locked?\n", __func__);
        return HAL_EEREAD;
    }
    HALDEBUG(ah, "EEPROM protect 0x%x\n", eeval);
    ahp->ah_eeprotect = eeval;
    /* XXX check proper access before continuing */
#endif

    eepdata = (u_int16_t *)(pEeprom);
    sum =0;
    len = (AH_MIN(pEeprom->baseEepHeader.length, sizeof(ar5416_eeprom_t))) / sizeof(u_int16_t);
    //len = sizeof(ar5416_eeprom_t) / sizeof(u_int16_t);
    HALDEBUG(ah,"%s Eeprom Version 0x%x.%x\n",__func__,owl_get_eep_ver(pEeprom),owl_get_eep_rev(pEeprom));
    for (i = 0; i < len; i++) {
        sum ^= *eepdata++;
    }

    /* Check CRC - Attach should fail on a bad checksum */
    if ((sum != 0xffff) || 
#if 0
    	(owl_get_eep_ver(pEeprom) != AR5416_EEP_VER) ||
    	(owl_get_eep_rev(pEeprom) < AR5416_EEP_NO_BACK_VER)) {
#else
	(owl_get_eep_ver(pEeprom) != AR5416_EEP_VER) ){
#endif
        HALDEBUG(ah, "Bad EEPROM checksum 0x%x or revision 0x%04x\n",
                 sum, owl_get_eep_ver(pEeprom));
        return HAL_EEBADSUM;
    }

#if AH_BYTE_ORDER == AH_BIG_ENDIAN
    if (!(pEeprom->baseEepHeader.eepMisc & AR5416_EEPMISC_BIG_ENDIAN)) {
#else 
    if (pEeprom->baseEepHeader.eepMisc & AR5416_EEPMISC_BIG_ENDIAN) {
#endif
    /*
     * TODO: add code so this inefficient hardware setup works across all 
     * compile types - possibly with a notice of an inefficient endian setup.
     */
        HALDEBUG(ah, "EEPROM Endianness is not native to driver\n");
        return HAL_EEBADSUM;
    }

    return HAL_OK;
}

static void
ar5416FillEmuEeprom(struct ath_hal *ahp)
{
    ar5416_eeprom_t *eep = (ar5416_eeprom_t *)(AH5416(ahp)->ah_5416eeprom);

    eep->baseEepHeader.version = AR5416_EEP_VER << 12;
    eep->baseEepHeader.macAddr[0] = 0x00;
    eep->baseEepHeader.macAddr[1] = 0x03;
    eep->baseEepHeader.macAddr[2] = 0x7F;
    eep->baseEepHeader.macAddr[3] = 0xBA;
    eep->baseEepHeader.macAddr[4] = 0xD0;
    eep->baseEepHeader.regDmn[0] = 0x10;
    eep->baseEepHeader.opCapFlags = AR5416_OPFLAGS_11G | AR5416_OPFLAGS_11A;
    eep->baseEepHeader.deviceCap = 0;
    eep->baseEepHeader.rfSilent = 0;
    eep->modalHeader[0].noiseFloorThreshCh[0] = -1;
    eep->modalHeader[1].noiseFloorThreshCh[0] = -1;

#ifdef EEPROM_DUMP
    ar5416EepromDump(ah, eep);
#endif
}

static HAL_BOOL
ar5416FillEeprom(struct ath_hal *ah)
{
    struct ath_hal_5416 *ahp = AH5416(ah);
    ar5416_eeprom_t *eep = (ar5416_eeprom_t *)ahp->ah_5416eeprom;
    u_int16_t *eep_data;
    int addr;

    HALDEBUG(ah,"%s eep = 0x%p\n",__func__,eep);
    eep_data = (u_int16_t *)eep;
    for (addr = 0; addr < sizeof(ar5416_eeprom_t) / sizeof(u_int16_t); addr++) {
        if (!owl_eeprom_read(ah, addr + owl_eep_start_loc, eep_data))
            return AH_FALSE;

        eep_data ++;
    }
#ifdef EEPROM_DUMP
   ar5416EepromDump(ah, eep);
#endif

    return AH_TRUE;
}

static u_int16_t 
ar5416EepromGetSpurChan(struct ath_hal *ah, u_int16_t i,HAL_BOOL is2GHz)
{ 
	ar5416_eeprom_t *eep = (ar5416_eeprom_t *)(AH5416(ah)->ah_5416eeprom);
	
	HALASSERT(i <  AR_EEPROM_MODAL_SPURS );
	return eep->modalHeader[is2GHz].spurChans[i].spurChan;
}

static HAL_BOOL
ar5416GetChipPowerLimits(struct ath_hal *ah, HAL_CHANNEL *chans, u_int32_t nchans)
{
    struct ath_hal_5416 *ahp = AH5416(ah);

    return ahp->ah_rfHal.getChipPowerLim(ah, chans, nchans);
}

/* Get Eeprom configuration information */
static HAL_BOOL
ar5416GetEepromCapabilityInfo(struct ath_hal *ah, u_int16_t *capField) 
{
    struct ath_hal_5416 *ahp = AH5416(ah);
    struct ath_hal_private *ahpriv = AH_PRIVATE(ah);
    HAL_CAPABILITIES *pCap = &ahpriv->ah_caps;
    u_int16_t eeval;
    u_int32_t i, sum;
	
    /* XXX How about the serial number ? */
    /* Read Reg Domain */
    ahp->ah_regdomain = ar5416EepromGet(ah, EEP_REG_0);

    /* Get RF Kill */
    ahpriv->ah_rfsilent = ar5416EepromGet(ah, EEP_RF_SILENT);
    if (ahpriv->ah_rfsilent & EEP_RFSILENT_ENABLED) {
    	ahpriv->ah_rfkillEnabled = AH_TRUE;
	pCap->halRfSilentSupport = AH_TRUE;
    };

    /* Read the capability EEPROM location */
    *capField = ar5416EepromGet(ah, EEP_OP_CAP);
	
    /* Construct wireless mode from EEPROM */
    pCap->halWirelessModes = 0;
    eeval = ar5416EepromGet(ah, EEP_OP_MODE);
    if (eeval & AR5416_OPFLAGS_11A) {
	        pCap->halWirelessModes |= (HAL_MODE_11A | HAL_MODE_11NA);
		ahp->ah_Amode = AH_TRUE;
    }

    if (eeval & AR5416_OPFLAGS_11G) {
    pCap->halWirelessModes |= (HAL_MODE_11G | HAL_MODE_11NG);
    		ahp->ah_Gmode = AH_TRUE;
		ahp->ah_Bmode = AH_TRUE;
    /* WAR for 20833 */
    pCap->halWirelessModes |= (HAL_MODE_11A | HAL_MODE_11NA);
    }
    
    /* No 32KHz crystal */
    ahp->ah_exist32kHzCrystal = AH_FALSE;
    
    /* Get MAC Address */
    sum = 0;
    for (i = 0; i < 3; i++) {
    	eeval = ar5416EepromGet(ah, AR_EEPROM_MAC(i));
	sum += eeval;
	ahp->ah_macaddr[2*i] = eeval >> 8;
	ahp->ah_macaddr[2*i + 1] = eeval & 0xff;
	}
	if (sum == 0 || sum == 0xffff*3) {
	        HALDEBUG(ah, "%s: mac address read failed: %s\n",
	            __func__, ath_hal_ether_sprintf(ahp->ah_macaddr));
	        return AH_FALSE;
    }
 
    return AH_TRUE;
}

static HAL_BOOL
ar5416GetEepromNoiseFloorThresh(struct ath_hal *ah, const HAL_CHANNEL_INTERNAL *chan, int16_t *nft)
{
	switch (chan->channelFlags & CHANNEL_ALL_NOTURBO) {
		case CHANNEL_A:
		case CHANNEL_A_HT20:
		case CHANNEL_A_HT40:
	                *nft = (int16_t) ar5416EepromGet(ah, EEP_NFTHRESH_5);
	                break;
		case CHANNEL_B:
		case CHANNEL_G:
		case CHANNEL_G_HT20:
		case CHANNEL_G_HT40:
	                *nft = (int16_t) ar5416EepromGet(ah, EEP_NFTHRESH_2);
	                break;
		default:
	                HALDEBUG(ah, "%s: invalid channel flags 0x%x\n",
	                        __func__, chan->channelFlags);
		return AH_FALSE;
	}
	 return AH_TRUE;
}
/*
 * Copy EEPROM Conformance Testing Limits contents 
 * into the allocated space
 */
/* USE CTLS from chain zero */ 
#define CTL_CHAIN	0 
static void
ar5416ReadEepromCTLInfo(struct ath_hal *ah)
{
	HAL_EEPROM *ee = &AH5416(ah)->ah_eeprom;
	RD_EDGES_POWER *rep = ee->ee_rdEdgesPower;
	ar5416_eeprom_t *pEeprom = (ar5416_eeprom_t *)(AH5416(ah)->ah_5416eeprom);
	u_int32_t	i,j;

	for (i = 0; (pEeprom->ctlIndex[i] != 0) && (i < AR5416_NUM_CTLS); i++) {
			for (j = 0; j < NUM_EDGES; j ++) {
			/* XXX Confirm this is the right thing to do when an invalid channel is stored */
				if (pEeprom->ctlData[i].ctlEdges[CTL_CHAIN][j].bChannel == AR5416_BCHAN_UNUSED) {
					rep[j].rdEdge = 0;
					rep[j].twice_rdEdgePower = 0;
					rep[j].flag = 0;
				}else{
					rep[j].rdEdge = fbin2freq(pEeprom->ctlData[i].ctlEdges[CTL_CHAIN][j].bChannel, 
					((pEeprom->ctlIndex[i] & CTL_MODE_M) != CTL_11A) );
					rep[j].twice_rdEdgePower = pEeprom->ctlData[i].ctlEdges[CTL_CHAIN][j].tPower;
					rep[j].flag = (pEeprom->ctlData[i].ctlEdges[CTL_CHAIN][j].flag != 0);
				}
				//HALDEBUG(ah,"%s ctl=0x%x edge=%u pwr=%u (%s)\n",__func__,pEeprom->ctlIndex[i],
				//			rep[j].rdEdge,rep[j].twice_rdEdgePower, rep[j].flag ? "MidBand" : "Edge");
			}
			rep += NUM_EDGES;
	}
	ee->ee_numCtls = i;
	HALDEBUG(ah,"%s Numctls = %u\n",__func__,i);
}
#ifdef EEPROM_DUMP
static void
ar5416PrintPowerPerRate(struct ath_hal *ah,int16_t pRatesPower[])
{
    const u_int8_t rateString[][16] = { "6mb OFDM", " 9mb OFDM", "12mb OFDM", "18mb OFDM",
                                   "24mb OFDM", "36mb OFDM", "48mb OFDM", "54mb OFDM",
                                   "1L   CCK ", "2L   CCK ", "2S   CCK ", "5.5L CCK ",
                                   "5.5S CCK ", "11L  CCK ", "11S  CCK ", "XR       ",
                                   "HT20mcs 0", "HT20mcs 1", "HT20mcs 2", "HT20mcs 3",
                                   "HT20mcs 4", "HT20mcs 5", "HT20mcs 6", "HT20mcs 7",
                                   "HT40mcs 0", "HT40mcs 1", "HT40mcs 2", "HT40mcs 3",
                                   "HT40mcs 4", "HT40mcs 5", "HT40mcs 6", "HT40mcs 7",
                                   "Dup CCK  ", "Dup OFDM ", "Ext CCK  ", "Ext OFDM ",
    };
    int i;

    for (i = 0; i < Ar5416RateSize; i+=4) {
        HALDEBUG(ah, " %s %3d.%1d dBm | %s %3d.%1d dBm | %s %3d.%1d dBm | %s %3d.%1d dBm\n",
            rateString[i], pRatesPower[i] / 2, (pRatesPower[i] % 2) * 5,
            rateString[i + 1], pRatesPower[i + 1] / 2, (pRatesPower[i + 1] % 2) * 5,
            rateString[i + 2], pRatesPower[i + 2] / 2, (pRatesPower[i + 2] % 2) * 5,
            rateString[i + 3], pRatesPower[i + 3] / 2, (pRatesPower[i + 3] % 2) * 5);
    }
}

/**************************************************************
 * ar5416EepromDump
 *
 * Produce a formatted dump of the EEPROM structure
 */
static void 
ar5416EepromDump(struct ath_hal *ah, ar5416_eeprom_t *ar5416Eep)
{
    u_int16_t          i, j, k, c, r;
    BASE_EEP_HEADER   *pBase = &(ar5416Eep->baseEepHeader);
    MODAL_EEP_HEADER  *pModal =AH_NULL;
    CAL_TARGET_POWER_HT  *pPowerInfo = AH_NULL;
    CAL_TARGET_POWER_LEG  *pPowerInfoLeg = AH_NULL;
    HAL_BOOL  legacyPowers = AH_FALSE;
    CAL_DATA_PER_FREQ *pDataPerChannel;
    HAL_BOOL            is2Ghz = 0;
    u_int8_t           noMoreSpurs;
    u_int8_t           *pChannel;
    u_int16_t          *pSpurData;
    u_int16_t          channelCount, channelRowCnt, vpdCount, rowEndsAtChan;
    u_int16_t          xpdGainMapping[] = {0, 1, 2, 4};
    u_int16_t          xpdGainValues[AR5416_NUM_PD_GAINS], numXpdGain = 0, xpdMask;
    u_int16_t          numPowers = 0, numRates = 0, ratePrint = 0, numChannels, tempFreq;
    u_int16_t          numTxChains = (u_int16_t)( ((pBase->txMask >> 2) & 1) +
        ((pBase->txMask >> 1) & 1) + (pBase->txMask & 1) );

    const static char *sRatePrint[3][8] = {
      {"     6-24     ", "      36      ", "      48      ", "      54      ", "", "", "", ""},
      {"       1      ", "       2      ", "     5.5      ", "      11      ", "", "", "", ""},
      {"   HT MCS0    ", "   HT MCS1    ", "   HT MCS2    ", "   HT MCS3    ",
       "   HT MCS4    ", "   HT MCS5    ", "   HT MCS6    ", "   HT MCS7    "},
    };

    const static char *sTargetPowerMode[7] = {
      "5G OFDM ", "5G HT20 ", "5G HT40 ", "2G CCK  ", "2G OFDM ", "2G HT20 ", "2G HT40 ",
    };

    const static char *sCtlType[] = {
        "[ 11A base mode ]",
        "[ 11B base mode ]",
        "[ 11G base mode ]",
        "[ INVALID       ]",
        "[ INVALID       ]",
        "[ 2G HT20 mode  ]",
        "[ 5G HT20 mode  ]",
        "[ 2G HT40 mode  ]",
        "[ 5G HT40 mode  ]",
    };

    if (!ar5416Eep) {
        return;
    }

    /* Print Header info */
    pBase = &(ar5416Eep->baseEepHeader);
    HALDEBUG(ah, "\n");
    HALDEBUG(ah, " =======================Header Information======================\n");
    HALDEBUG(ah, " |  Major Version           %2d  |  Minor Version           %2d  |\n",
             pBase->version >> 12, pBase->version & 0xFFF);
    HALDEBUG(ah, " |-------------------------------------------------------------|\n");
    HALDEBUG(ah, " |  Checksum           0x%04X   |  Length             0x%04X   |\n",
             pBase->checksum, pBase->length);
    HALDEBUG(ah, " |  RegDomain 1        0x%04X   |  RegDomain 2        0x%04X   |\n",
             pBase->regDmn[0], pBase->regDmn[1]);
    HALDEBUG(ah, " |  MacAddress: 0x%02X:%02X:%02X:%02X:%02X:%02X                            |\n",
             pBase->macAddr[0], pBase->macAddr[1], pBase->macAddr[2],
             pBase->macAddr[3], pBase->macAddr[4], pBase->macAddr[5]);
    HALDEBUG(ah, " |  TX Mask            0x%04X   |  RX Mask            0x%04X   |\n",
             pBase->txMask, pBase->rxMask);
    HALDEBUG(ah, " |  OpFlags: 11A %d, 11G %d                                      |\n",
             (pBase->opCapFlags & AR5416_OPFLAGS_11A) || 0, (pBase->opCapFlags & AR5416_OPFLAGS_11G) || 0);

    HALDEBUG(ah, " |  Customer Data in hex                                       |\n");
    for (i = 0; i < 64; i++) {
        if ((i % 16) == 0) {
            HALDEBUG(ah, " |= ");
        }
        HALDEBUG(ah, "%02X ", ar5416Eep->custData[i]);
        if ((i % 16) == 15) {
            HALDEBUG(ah, "          =|\n");
        }
    }

    /* Print Modal Header info */
    for (i = 0; i < 2; i++) {
        if ((i == 0) && !(pBase->opCapFlags & AR5416_OPFLAGS_11A)) {
            continue;
        }
        if ((i == 1) && !(pBase->opCapFlags & AR5416_OPFLAGS_11G)) {
            continue;
        }
        pModal = &(ar5416Eep->modalHeader[i]);
        if (i == 0) {
            HALDEBUG(ah, " |======================5GHz Modal Header======================|\n");
        } else {
            HALDEBUG(ah, " |======================2GHz Modal Header======================|\n");
        }
        HALDEBUG(ah, " |  Ant Chain 0     0x%08X  |  Ant Chain 1     0x%08X  |\n",
                 pModal->antCtrlChain[0], pModal->antCtrlChain[1]);
        HALDEBUG(ah, " |  Ant Chain 2     0x%08X  |\n",
                 pModal->antCtrlChain[2]);
        HALDEBUG(ah, " |  Antenna Common  0x%08X  |  Antenna Gain Chain 0   %3d  |\n",
                 pModal->antCtrlCommon, pModal->antennaGainCh[0]);
        HALDEBUG(ah, " |  Antenna Gain Chain 1   %3d  |  Antenna Gain Chain 2   %3d  |\n",
                 pModal->antennaGainCh[1], pModal->antennaGainCh[2]);
        HALDEBUG(ah, " |  Switch Settling        %3d  |  TxRxAttenuation Ch 0   %3d  |\n",
                 pModal->switchSettling, pModal->txRxAttenCh[0]);
        HALDEBUG(ah, " |  TxRxAttenuation Ch 1   %3d  |  TxRxAttenuation Ch 2   %3d  |\n",
                 pModal->txRxAttenCh[1], pModal->txRxAttenCh[2]);
        HALDEBUG(ah, " |  RxTxMargin Chain 0     %3d  |  RxTxMargin Chain 1     %3d  |\n",
                 pModal->rxTxMarginCh[0], pModal->rxTxMarginCh[1]);
        HALDEBUG(ah, " |  RxTxMargin Chain 2     %3d  |  adc desired size       %3d  |\n",
                 pModal->rxTxMarginCh[2], pModal->adcDesiredSize);
        HALDEBUG(ah, " |  pga desired size       %3d  |  tx end to xlna on      %3d  |\n",
                 pModal->pgaDesiredSize, pModal->txEndToRxOn);
        HALDEBUG(ah, " |  xlna gain Chain 0      %3d  |  xlna gain Chain 1      %3d  |\n",
                 pModal->xlnaGainCh[0], pModal->xlnaGainCh[1]);
        HALDEBUG(ah, " |  xlna gain Chain 2      %3d  |  tx end to xpa off      %3d  |\n",
                 pModal->xlnaGainCh[2], pModal->txEndToXpaOff);
        HALDEBUG(ah, " |  tx frame to xpa on     %3d  |  thresh62               %3d  |\n",
                 pModal->txFrameToXpaOn, pModal->thresh62);
        HALDEBUG(ah, " |  noise floor thres 0    %3d  |  noise floor thres 1    %3d  |\n",
                 pModal->noiseFloorThreshCh[0], pModal->noiseFloorThreshCh[1]);
        HALDEBUG(ah, " |  noise floor thres 2    %3d  |  Xpd Gain Mask 0x%X | Xpd %2d  |\n",
                 pModal->noiseFloorThreshCh[2], pModal->xpdGain, pModal->xpd);
        HALDEBUG(ah, " |  IQ Cal I Chain 0       %3d  |  IQ Cal Q Chain 0       %3d  |\n",
                 pModal->iqCalICh[0], pModal->iqCalQCh[0]);
        HALDEBUG(ah, " |  IQ Cal I Chain 1       %3d  |  IQ Cal Q Chain 1       %3d  |\n",
                 pModal->iqCalICh[1], pModal->iqCalQCh[1]);
        HALDEBUG(ah, " |  IQ Cal I Chain 2       %3d  |  IQ Cal Q Chain 2       %3d  |\n",
                 pModal->iqCalICh[2], pModal->iqCalQCh[2]);
    }
    HALDEBUG(ah, " |=============================================================|\n");

#define SPUR_TO_KHZ(is2GHz, spur)    ( ((spur) + ((is2GHz) ? 23000 : 49000)) * 100 )
#define NO_SPUR                      ( 0x8000 )

    /* Print spur data */
    HALDEBUG(ah, "=======================Spur Information======================\n");
    for (i = 0; i < 2; i++) {
        if ((i == 0) && !(pBase->opCapFlags & AR5416_OPFLAGS_11A)) {
            continue;
        }
        if ((i == 1) && !(pBase->opCapFlags & AR5416_OPFLAGS_11G)) {
            continue;
        }
        if (i == 0) {
            HALDEBUG(ah, "| 11A Spurs in MHz                                          |\n");
        } else {
            HALDEBUG(ah, "| 11G Spurs in MHz                                          |\n");
        }
        pSpurData = (u_int16_t *)ar5416Eep->modalHeader[i].spurChans;
        noMoreSpurs = 0;
        for (j = 0; j < AR5416_EEPROM_MODAL_SPURS; j++) {
            if ((pModal->spurChans[j].spurChan == NO_SPUR) || noMoreSpurs) {
                noMoreSpurs = 1;
                HALDEBUG(ah, "|  NO SPUR  ");
            } else {
                HALDEBUG(ah, "|   %4d.%1d  ", SPUR_TO_KHZ(i,pModal->spurChans[j].spurChan )/1000,
                         ((SPUR_TO_KHZ(i, pModal->spurChans[j].spurChan))/100) % 10);
            }
        }
	for (j = 0; j < AR5416_EEPROM_MODAL_SPURS; j++) {
            if ((pModal->spurChans[j].spurChan == NO_SPUR) || noMoreSpurs) {
                noMoreSpurs = 1;
                HALDEBUG(ah,"|             ");
            } else {
                HALDEBUG(ah,"|<%2d.%1d-=-%2d.%1d>", 
                       pModal->spurChans[j].spurRangeLow / 10, pModal->spurChans[j].spurRangeLow % 10,
                       pModal->spurChans[j].spurRangeHigh / 10, pModal->spurChans[j].spurRangeHigh % 10);
            }
        }
        HALDEBUG(ah, "|\n");
    }
    HALDEBUG(ah, "|===========================================================|\n");

#undef SPUR_TO_KHZ
#undef NO_SPUR

    /* Print calibration info */
    for (i = 0; i < 2; i++) {
        if ((i == 0) && !(pBase->opCapFlags & AR5416_OPFLAGS_11A)) {
            continue;
        }
        if ((i == 1) && !(pBase->opCapFlags & AR5416_OPFLAGS_11G)) {
            continue;
        }
        for (c = 0; c < AR5416_MAX_CHAINS; c++) {
            if (!((pBase->txMask >> c) & 1)) {
                continue;
            }
            if (i == 0) {
                pDataPerChannel = ar5416Eep->calPierData5G[c];
                numChannels = AR5416_NUM_5G_CAL_PIERS;
                pChannel = ar5416Eep->calFreqPier5G;
            } else {
                pDataPerChannel = ar5416Eep->calPierData2G[c];
                numChannels = AR5416_NUM_2G_CAL_PIERS;
                pChannel = ar5416Eep->calFreqPier2G;
            }
            xpdMask = ar5416Eep->modalHeader[i].xpdGain;

            numXpdGain = 0;
            /* Calculate the value of xpdgains from the xpdGain Mask */
            for (j = 1; j <= AR5416_PD_GAINS_IN_MASK; j++) {
                if ((xpdMask >> (AR5416_PD_GAINS_IN_MASK - j)) & 1) {
                    if (numXpdGain >= AR5416_NUM_PD_GAINS) {
                        HALASSERT(0);
                        break;
                    }
                    xpdGainValues[numXpdGain++] = (u_int16_t)(AR5416_PD_GAINS_IN_MASK - j);
                }
            }

            HALDEBUG(ah, "===============Power Calibration Information Chain %d =======================\n", c);
            for (channelRowCnt = 0; channelRowCnt < numChannels; channelRowCnt += 5) {
                rowEndsAtChan = (u_int16_t)AH_MIN(numChannels, (channelRowCnt + 5));
                for (channelCount = channelRowCnt; channelCount < rowEndsAtChan; channelCount++) {
                    HALDEBUG(ah, "|     %04d     ", fbin2freq(pChannel[channelCount], (HAL_BOOL)i));
                }
                HALDEBUG(ah, "|\n|==============|==============|==============|==============|==============|\n");
                for (channelCount = channelRowCnt; channelCount < rowEndsAtChan; channelCount++) {
                    HALDEBUG(ah, "|pdadc pwr(dBm)");
                }
                HALDEBUG(ah, "|\n");

                for (k = 0; k < numXpdGain; k++) {
                    HALDEBUG(ah, "|              |              |              |              |              |\n");
                    HALDEBUG(ah, "| PD_Gain %2d   |              |              |              |              |\n",
                             xpdGainMapping[xpdGainValues[k]]);

                    for (vpdCount = 0; vpdCount < AR5416_PD_GAIN_ICEPTS; vpdCount++) {
                        for (channelCount = channelRowCnt; channelCount < rowEndsAtChan; channelCount++) {
                            HALDEBUG(ah, "|  %02d   %2d.%02d  ", pDataPerChannel[channelCount].vpdPdg[k][vpdCount],
                                     pDataPerChannel[channelCount].pwrPdg[k][vpdCount] / 4,
                                     (pDataPerChannel[channelCount].pwrPdg[k][vpdCount] % 4) * 25);
                        }
                        HALDEBUG(ah, "|\n");
                    }
                }
                HALDEBUG(ah, "|              |              |              |              |              |\n");
                HALDEBUG(ah, "|==============|==============|==============|==============|==============|\n");
            }
            HALDEBUG(ah, "|\n");
        }
    }

    /* Print Target Powers */
    for (i = 0; i < 7; i++) {
        if ( (i <=2) && !(pBase->opCapFlags & AR5416_OPFLAGS_11A)) {
            continue;
        }
        if ((i >= 3 && i <=6) && !(pBase->opCapFlags & AR5416_OPFLAGS_11G)) {
            continue;
        }
        switch(i) {
        case 0:
            pPowerInfo = (CAL_TARGET_POWER_HT *)ar5416Eep->calTargetPower5G;
            pPowerInfoLeg = ar5416Eep->calTargetPower5G;
            numPowers = AR5416_NUM_5G_20_TARGET_POWERS;
            ratePrint = 0;
            numRates = 4;
            is2Ghz = AH_FALSE;
            legacyPowers = AH_TRUE;
            break;
        case 1:
            pPowerInfo = ar5416Eep->calTargetPower5GHT20;
            numPowers = AR5416_NUM_5G_20_TARGET_POWERS;
            ratePrint = 2;
            numRates = 8;
            is2Ghz = AH_FALSE;
            legacyPowers = AH_FALSE;
            break;
        case 2:
            pPowerInfo = ar5416Eep->calTargetPower5GHT40;
            numPowers = AR5416_NUM_5G_40_TARGET_POWERS;
            ratePrint = 2;
            numRates = 8;
            is2Ghz = AH_FALSE;
            legacyPowers = AH_FALSE;
            break;
        case 3:
            pPowerInfo = (CAL_TARGET_POWER_HT *)ar5416Eep->calTargetPowerCck;
            pPowerInfoLeg = ar5416Eep->calTargetPowerCck;
            numPowers = AR5416_NUM_2G_CCK_TARGET_POWERS;
            ratePrint = 1;
            numRates = 4;
            is2Ghz = AH_TRUE;
            legacyPowers = AH_TRUE;
            break;
        case 4:
            pPowerInfo = (CAL_TARGET_POWER_HT *)ar5416Eep->calTargetPower2G;
            pPowerInfoLeg = ar5416Eep->calTargetPower2G;
            numPowers = AR5416_NUM_2G_20_TARGET_POWERS;
            ratePrint = 0;
            numRates = 4;
            is2Ghz = AH_TRUE;
            legacyPowers = AH_TRUE;
           break;
        case 5:
            pPowerInfo = ar5416Eep->calTargetPower2GHT20;
            numPowers = AR5416_NUM_2G_20_TARGET_POWERS;
            ratePrint = 2;
            numRates = 8;
            is2Ghz = AH_TRUE;
            legacyPowers = AH_FALSE;
            break;
        case 6:
            pPowerInfo = ar5416Eep->calTargetPower2GHT40;
            numPowers = AR5416_NUM_2G_40_TARGET_POWERS;
            ratePrint = 2;
            numRates = 8;
            is2Ghz = AH_TRUE;
            legacyPowers = AH_FALSE;
            break;
        }
        HALDEBUG(ah, "============================Target Power Info===============================\n");
        for (j = 0; j < numPowers; j+=4) {
            HALDEBUG(ah, "|   %s   ", sTargetPowerMode[i]);
            for (k = j; k < AH_MIN(j + 4, numPowers); k++) {
                if(legacyPowers) {
                    HALDEBUG(ah, "|     %04d     ",fbin2freq(pPowerInfoLeg[k].bChannel, (HAL_BOOL)i));
                } else {
                    HALDEBUG(ah, "|     %04d     ",fbin2freq(pPowerInfo[k].bChannel, (HAL_BOOL)i));
                }
            }
            HALDEBUG(ah, "|\n");
            HALDEBUG(ah, "|==============|==============|==============|==============|==============|\n");
            for (r = 0; r < numRates; r++) {
                HALDEBUG(ah, "|%s", sRatePrint[ratePrint][r]);
                for (k = j; k < AH_MIN(j + 4, numPowers); k++) {
                    if(legacyPowers) {
                        HALDEBUG(ah, "|     %2d.%d     ", pPowerInfoLeg[k].tPow2x[r] / 2,
                             (pPowerInfo[k].tPow2x[r] % 2) * 5);
                    } else {
                        HALDEBUG(ah, "|     %2d.%d     ", pPowerInfo[k].tPow2x[r] / 2,
                             (pPowerInfo[k].tPow2x[r] % 2) * 5);
                    }
                }
                HALDEBUG(ah, "|\n");
            }
            HALDEBUG(ah, "|==============|==============|==============|==============|==============|\n");
        }
    }
    HALDEBUG(ah, "\n");

    /* Print Band Edge Powers */
    HALDEBUG(ah, "=======================Test Group Band Edge Power========================\n");
    for (i = 0; (ar5416Eep->ctlIndex[i] != 0) && (i < AR5416_NUM_CTLS); i++) {
        HALDEBUG(ah, "|                                                                       |\n");
        HALDEBUG(ah, "| CTL: 0x%02x %s                                           |\n",
                 ar5416Eep->ctlIndex[i], sCtlType[ar5416Eep->ctlIndex[i] & CTL_MODE_M]);
        HALDEBUG(ah, "|=======|=======|=======|=======|=======|=======|=======|=======|=======|\n");

        for (c = 0; c < numTxChains; c++) {
            HALDEBUG(ah, "======================== Edges for Chain %d ==============================\n", c);
            HALDEBUG(ah, "| edge  ");
            for (j = 0; j < AR5416_NUM_BAND_EDGES; j++) {
                if (ar5416Eep->ctlData[i].ctlEdges[c][j].bChannel == AR5416_BCHAN_UNUSED) {
                    HALDEBUG(ah, "|  --   ");
                } else {
                    tempFreq = fbin2freq(ar5416Eep->ctlData[i].ctlEdges[c][j].bChannel, is2Ghz);
                    HALDEBUG(ah, "| %04d  ", tempFreq);
                }
            }
            HALDEBUG(ah, "|\n");
            HALDEBUG(ah, "|=======|=======|=======|=======|=======|=======|=======|=======|=======|\n");

            HALDEBUG(ah, "| power ");
            for (j = 0; j < AR5416_NUM_BAND_EDGES; j++) {
                if (ar5416Eep->ctlData[i].ctlEdges[c][j].bChannel == AR5416_BCHAN_UNUSED) {
                    HALDEBUG(ah, "|  --   ");
                } else {
                    HALDEBUG(ah, "| %2d.%d  ", ar5416Eep->ctlData[i].ctlEdges[c][j].tPower / 2,
                        (ar5416Eep->ctlData[i].ctlEdges[c][j].tPower % 2) * 5);
                }
            }
            HALDEBUG(ah, "|\n");
            HALDEBUG(ah, "|=======|=======|=======|=======|=======|=======|=======|=======|=======|\n");

            HALDEBUG(ah, "| flag  ");
            for (j = 0; j < AR5416_NUM_BAND_EDGES; j++) {
                if (ar5416Eep->ctlData[i].ctlEdges[c][j].bChannel == AR5416_BCHAN_UNUSED) {
                    HALDEBUG(ah, "|  --   ");
                } else {
                    HALDEBUG(ah, "|   %1d   ", ar5416Eep->ctlData[i].ctlEdges[c][j].flag);
                }
            }
            HALDEBUG(ah, "|\n");
            HALDEBUG(ah, "=========================================================================\n");
        }
    }

}
#undef N
#undef POW_SM
#endif
#endif /* AH_SUPPORT_AR5416 */
